
# coding: utf-8

# In[4]:

def selection_genetica(target,df,modelo,train_size):
    
    from deap import creator, base, tools, algorithms #GENETIC ALGORITHM LIBRARY - requirement: pip install deap
    import random
    from sklearn import metrics,linear_model,svm,lda,cross_validation
    import pandas as pd
    import re 
    import numpy as np
    import statsmodels.api as sm
    import string

    # VARIABLES

    output_var = target

    list_inputs = set(df.drop(output_var,1).columns)



    print "GENETIC ALGORITHM FOR FEATURE SELECTION:"

    #####
    #SETING UP THE GENETIC ALGORITHM and CALCULATING STARTING POOL (STARTING CANDIDATE POPULATION)
    #####
    creator.create("FitnessMax", base.Fitness, weights=(1.0,))
    creator.create("Individual", list, fitness=creator.FitnessMax)
    toolbox = base.Toolbox()
    toolbox.register("attr_bool", random.randint, 0, 1)
    toolbox.register("individual", tools.initRepeat, creator.Individual, toolbox.attr_bool, n=len(list_inputs))
    toolbox.register("population", tools.initRepeat, list, toolbox.individual)
    def evalOneMax(individual):
        return sum(individual),

    toolbox.register("evaluate", evalOneMax)
    toolbox.register("mate", tools.cxTwoPoint)
    toolbox.register("mutate", tools.mutFlipBit, indpb=0.05)
    toolbox.register("select", tools.selTournament, tournsize=3)



    NPOPSIZE = 50 #RANDOM STARTING POOL SIZE
    population = toolbox.population(n=NPOPSIZE)



    #####
    #ASSESSING GINI ON THE STARTING POOL
    #####
    dic_gini={}
    for i in range(np.shape(population)[0]): 

        # TRASLATING DNA INTO LIST OF VARIABLES (1-81)
        var_model = []    
        for j in range(np.shape(population)[1]): 
            if (population[i])[j]==1:
                var_model.append(list(list_inputs)[j])

        # ASSESSING GINI INDEX FOR EACH INVIVIDUAL IN THE INITIAL POOL 
        df_train,df_test=cross_validation.train_test_split(df,train_size=train_size)     
        X_train=df_train[var_model]
        Y_train=df_train[output_var]
        X_test=df_test[var_model]
        Y_test=df_test[output_var]
        X_train=X_train.values
        X_test=X_test.values
        ######
        # CHANGE_HERE - START: YOU ARE VERY LIKELY USING A DIFFERENT TECHNIQUE BY NOW. SO CHANGE TO YOURS.
        #####   

        if modelo=="LogisticRegression":
            clf=linear_model.LogisticRegression()

        elif modelo==("SVC"):

            clf=svm.SVC(probability=True)



        elif modelo=="LDA":
            clf = lda.LDA()

        clf.fit(X_train,Y_train)
        Y_predict=np.transpose(clf.predict_proba(X_test))[1]

        ######
        # CHANGE_HERE - END: YOU ARE VERY LIKELY USING A DIFFERENT TECHNIQUE BY NOW. SO CHANGE TO YOURS.
        #####             


        ######
        # CHANGE_HERE - START: HERE IT USES THE DEVELOPMENT GINI TO SELECT VARIABLES, YOU SHOULD A DIFFERENT GINI. EITHER THE OOT GINI OR THE SQRT(DEV_GINI*OOT_GINI)
        #####                
        fpr, tpr, thresholds = metrics.roc_curve(Y_test, Y_predict)
        auc = metrics.auc(fpr, tpr)
        gini_power = abs(2*auc-1)
        ######
        # CHANGE_HERE - END: HERE IT USES THE DEVELOPMENT GINI TO SELECT VARIABLES, YOU SHOULD A DIFFERENT GINI. EITHER THE OOT GINI OR THE SQRT(DEV_GINI*OOT_GINI)
        #####                

        gini=str(gini_power)+";"+str(population[i]).replace('[','').replace(', ','').replace(']','')
        dic_gini[gini]=population[i]   
    list_gini=sorted(dic_gini.keys(),reverse=True)



    #####
    #GENETIC ALGORITHM MAIN LOOP - START
    # - ITERATING MANY TIMES UNTIL NO IMPROVMENT HAPPENS IN ORDER TO FIND THE OPTIMAL SET OF CHARACTERISTICS (VARIABLES)
    #####
    sum_current_gini=0.0
    sum_current_gini_1=0.0
    sum_current_gini_2=0.0
    first=0    
    OK = 1
    a=0

    while OK:  #REPEAT UNTIL IT DO NOT IMPROVE, AT LEAST A LITLE, THE GINI IN 2 GENERATIONS
        a=a+1
        print 'loop ', a
        OK=0
        ####
        # GENERATING OFFSPRING - START
        ####

        offspring = algorithms.varAnd(population, toolbox, cxpb=0.5, mutpb=0.1) #CROSS-X PROBABILITY = 50%, MUTATION PROBABILITY=10%

        fits = toolbox.map(toolbox.evaluate, offspring)
        for fit, ind in zip(fits, offspring):
            ind.fitness.values = fit
        population =toolbox.select(offspring, k=len(population))
        ####
        # GENERATING OFFSPRING - END
        ####

        sum_current_gini_2=sum_current_gini_1
        sum_current_gini_1=sum_current_gini
        sum_current_gini=0.0

        #####
        #ASSESSING GINI ON THE OFFSPRING - START
        #####
        for j in range(np.shape(population)[0]): 
            if population[j] not in dic_gini.values(): 
                var_model = [] 
                for i in range(np.shape(population)[1]): 
                    if (population[j])[i]==1:
                        var_model.append(list(list_inputs)[i])

                # CREACCION CONJUNTO TRAIN Y TEST

                df_train,df_test=cross_validation.train_test_split(df,train_size=train_size)     
                X_train=df_train[var_model]
                Y_train=df_train[output_var]
                X_test=df_test[var_model]
                Y_test=df_test[output_var]
                X_train=X_train.values
                X_test=X_test.values

                ######
                # CHANGE_HERE - START: YOU ARE VERY LIKELY USING A DIFFERENT TECHNIQUE BY NOW. SO CHANGE TO YOURS.
                #####   

                # diferentes modelos según la opción elegida

                if modelo=="LogisticRegression":
                    clf=linear_model.LogisticRegression()

                elif modelo==("SVC"):

                    clf=svm.SVC(probability=True)


                elif modelo=="LDA":
                    clf = lda.LDA()


                clf.fit(X_train,Y_train)
                Y_predict=np.transpose(clf.predict_proba(X_test))[1]
                ######
                # CHANGE_HERE - END: YOU ARE VERY LIKELY USING A DIFFERENT TECHNIQUE BY NOW. SO CHANGE TO YOURS.
                #####            


                ######
                # CHANGE_HERE - START: HERE IT USES THE DEVELOPMENT GINI TO SELECT VARIABLES, YOU SHOULD A DIFFERENT GINI. EITHER THE OOT GINI OR THE SQRT(DEV_GINI*OOT_GINI)
                #####                       
                fpr, tpr, thresholds = metrics.roc_curve(Y_test, Y_predict)
                auc = metrics.auc(fpr, tpr)
                gini_power = abs(2*auc-1)
                ######
                # CHANGE_HERE - END: HERE IT USES THE DEVELOPMENT GINI TO SELECT VARIABLES, YOU SHOULD A DIFFERENT GINI. EITHER THE OOT GINI OR THE SQRT(DEV_GINI*OOT_GINI)
                #####                       

                gini=str(gini_power)+";"+str(population[j]).replace('[','').replace(', ','').replace(']','')
                dic_gini[gini]=population[j]  
        #####
        #ASSESSING GINI ON THE OFFSPRING - END
        #####

        #####
        #SELECTING THE BEST FITTED AMONG ALL EVER CREATED POPULATION AND CURRENT OFFSPRING - START
        #####           
        list_gini=sorted(dic_gini.keys(),reverse=True)
        population=[]
        for i in list_gini[:NPOPSIZE]:
            population.append(dic_gini[i])
            gini=float(i.split(';')[0])
            sum_current_gini+=gini


        #####
        #SELECTING THE BEST FITTED AMONG ALL EVER CREATED POPULATION AND CURRENT OFFSPRING - END
        #####           

        #HAS IT IMPROVED AT LEAST A LITLE THE GINI IN THE LAST 2 GENERATIONS
        print 'sum_current_gini=', sum_current_gini, 'sum_current_gini_1=', sum_current_gini_1, 'sum_current_gini_2=', sum_current_gini_2
        if(sum_current_gini>sum_current_gini_1+0.0001 or sum_current_gini>sum_current_gini_2+0.0001):
            OK=1
    #####
    #GENETIC ALGORITHM MAIN LOOP - END
    #####


    gini_max=list_gini[0]        
    gini=float(gini_max.split(';')[0])
    features=gini_max.split(';')[1]


    ####
    # PRINTING OUT THE LIST OF FEATURES
    #####
    f=0
    for i in range(len(features)):
        if features[i]=='1':
            f+=1
            print 'feature ', f, ':', list(list_inputs)[i]
    print 'gini: ', gini

    lista_var=[]
    f=0
    for i in range(len(features)):
        if features[i]=='1':
            f+=1
            lista_var.append(list(list_inputs)[i])

    return lista_var


# In[5]:

# Aplicado a dataframe

def modelos(target,data,train_size):
    
    from sklearn import metrics
    from sklearn import linear_model,svm,cross_validation,tree,lda,ensemble
    import pandas as pd
    import numpy as np
    
    
    seed=284772184
    lista_modelos=["LogisticRegression","SVC","LDA","DecisionTreeClassifier","RandomForestClassifier"]
    lista_auc=[]
    profundidades=[1,2,3,4]
    
    for modelo in lista_modelos:
        print "CALCULO DE AUC PARA EL MODELO:",modelo
        lista_auc_tree=[]
        
        if modelo in ("LogisticRegression","LDA","SVC") :
            
            if modelo=="LogisticRegression":
                clf=linear_model.LogisticRegression()
            
            elif modelo==("SVC"):

                clf=svm.SVC(probability=True)


            else:
                clf=lda.LDA()

            
            lista_var=selection_genetica(target,data,modelo,train_size)
            
            lista_var.append(target)
            data2=data[lista_var]
            
            data_train,data_test=cross_validation.train_test_split(data2,train_size=train_size,random_state=seed)
        
            Y_train=data_train[target]
            X_train=data_train.drop(target,1)

            Y_test=data_test[target]
            X_test=data_test.drop(target,1)
             

            
            clf.fit(X_train,Y_train)

            Y_predict=np.transpose(clf.predict_proba(X_test))[1]
            
            fpr, tpr, thresholds = metrics.roc_curve(Y_test, Y_predict)
            auc = metrics.auc(fpr, tpr)
            lista_auc.append(auc)
        
        elif modelo in ("DecisionTreeClassifier","RandomForestClassifier"):
            
            if modelo=="DecisionTreeClassifier":
                clf = tree.DecisionTreeClassifier()
            
            else:
                clf = ensemble.RandomForestClassifier()
            
            
            data_train,data_test=cross_validation.train_test_split(data,train_size=train_size,random_state=seed)
            
            Y_train=data_train[target]
            X_train=data_train.drop(target,1)

            Y_test=data_test[target]
            X_test=data_test.drop(target,1)
            
            
            for d in profundidades:

                # ENTRENAMIENTO
                
                clf.set_params(max_depth=d)
                clf.fit(X_train,Y_train)


                # PREDICTION EN EL TEST
                Y_predict=np.transpose(clf.predict_proba(X_test))[1]
                fpr, tpr, thresholds = metrics.roc_curve(Y_test, Y_predict)
                auc = metrics.auc(fpr, tpr)
                
                lista_auc_tree.append(auc)
                
            lista_auc.append(np.min(lista_auc_tree))
                


    max1_pos=np.argmax(lista_auc)
    max1=np.max(lista_auc)
    lista_auc[max1_pos]=-2732
    max2_pos=np.argmax(lista_auc)
    max2=np.max(lista_auc)
    lista_auc[max2_pos]=-2732
    max3_pos=np.argmax(lista_auc)
    max3=np.max(lista_auc)
    lista_auc[max3_pos]=-2732
    max4_pos=np.argmax(lista_auc)
    max4=np.max(lista_auc)
    lista_auc[max4_pos]=-2732
    max5_pos=np.argmax(lista_auc)
    max5=np.max(lista_auc)
    

    print "\n"
    print "\n"
    print "Primero modelo con mayor auc:",lista_modelos[max1_pos],"con auc de:",max1
    print "Segundo modelo con mayor auc:",lista_modelos[max2_pos],"con auc de:",max2
    print "Tercero modelo con mayor auc:",lista_modelos[max3_pos],"con auc de:",max3
    print "Cuarto modelo con mayor auc:",lista_modelos[max4_pos],"con auc de:",max4
    print "Quinto modelo con mayor auc:",lista_modelos[max5_pos],"con auc de:",max5
            




"""ResNet models for Keras.

# Reference paper

- [Deep Residual Learning for Image Recognition]
  (https://arxiv.org/abs/1512.03385) (CVPR 2016 Best Paper Award)

# Reference implementations

- [TensorNets]
  (https://github.com/taehoonlee/tensornets/blob/master/tensornets/resnets.py)
- [Caffe ResNet]
  (https://github.com/KaimingHe/deep-residual-networks/tree/master/prototxt)

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from . import imagenet_utils
from .imagenet_utils import decode_predictions
from .resnet_common import ResNet50
from .resnet_common import ResNet101
from .resnet_common import ResNet152


def preprocess_input(x, **kwargs):
    """Preprocesses a numpy array encoding a batch of images.

    # Arguments
        x: a 4D numpy array consists of RGB values within [0, 255].
        data_format: data format of the image tensor.

    # Returns
        Preprocessed array.
    """
    return imagenet_utils.preprocess_input(x, mode='caffe', **kwargs)



# -*- coding: utf-8 -*-
"""
Created on Thu Jul 23 16:40:27 2020

@author: Koerber
"""

import cv2
import numpy as np
import tensorflow as tf
from abc import ABC, abstractmethod
import dl.training.datagenerator as datagenerator
from tensorflow.keras.models import load_model
import os
from dl.models.segmentation_models import Unet, FPN, Linknet, PSPNet, get_preprocessing
from dl.models.deeplab.deeplab import Deeplabv3, deeplab_preprocess, deeplab_backbones
from dl.models.segmentation_models.backbones.backbones_factory import Backbones

from enum import Enum


class PixelBasedPrediction(ABC):
    def __init__(self, parent):
        self.parent = parent

    def PredictImage(self, image):    
        if not self.parent.initialized or image is None:
            return None


        width = image.shape[1]
        height = image.shape[0]
        image = cv2.resize(image, (int(width*self.parent.ImageScaleFactor), int(height*self.parent.ImageScaleFactor)))

        pred = self.PredictFullImage(image)
        pred = self.resizeLabel(pred, (width, height))
        return pred

    def PredictFullImage(self, image):
        if image.shape[0] * image.shape[1] < self.parent.augmentation.outputwidth * self.parent.augmentation.outputheight:
            return self.Predict(image)
        
        results = []
        origImage = image
        
        res = self.PredictFromGenerator(origImage)

        results.append(res)
        
        if self.parent.tta:
            for i in range(5):               
                if i < 2:
                    flip = i
                    res = self.PredictFromGenerator(np.flip(origImage.copy(),flip))
                    results.append(np.flip(res,flip))
                else:
                    axes = (0,1)
                    iaxes = (1,0)
                    res = self.PredictFromGenerator(np.rot90(origImage.copy(), k = i-1, axes=axes))
                    results.append(np.rot90(res, k = i-1, axes=iaxes))


        results = np.stack(results, axis = 0)
        result = np.mean(results, axis = 0)
        return self.convert2Image(result).astype('uint8')

    def PredictFromGenerator(self, image):       
        generator = datagenerator.PredictionDataGenerator(self.parent, image)
        pred = self.parent.Model.predict(generator, workers=self.parent.worker)
        result = np.zeros(image.shape[0:2] + (self.parent.NumClasses,))
        sum_result = np.zeros(result.shape)
        
        border = self.parent.ModelOverlap//2
        p_width = self.parent.augmentation.outputwidth - 2*border
        p_height = self.parent.augmentation.outputheight - 2*border
        
        grad_patch = self.createModelOutputGradient()

        i = 0
        for h in range (0,image.shape[0],p_height):
            for w in range (0,image.shape[1],p_width):
                h_end = min(h+p_height,image.shape[0])
                w_end = min(w+p_width,image.shape[1])
                h_0 = min(h, result.shape[0]-p_height)
                w_0 = min(w, result.shape[1]-p_width)
        
                patch = pred[i]
                i += 1
                ### calculate the overlap with the help of grad_patch
                real_h0 = max(h_0-border,0)
                real_h_end = min(h_end+border,result.shape[0])
                real_w0 = max(w_0-border,0)
                real_w_end = min(w_end+border,result.shape[1])
                
                ph_0 = 0
                ph_end = patch.shape[0]
                pw_0 = 0
                pw_end = patch.shape[1]
                
                if real_h0 != h_0-border:
                    ph_0 = border - h_0
                if real_h_end != h_end+border:
                    ph_end = patch.shape[0] - (h_end+border - result.shape[0])
                
                if real_w0 != w_0-border:
                    pw_0 = border - w_0
                if real_w_end != w_end+border:
                    pw_end = patch.shape[1] - (w_end+border - result.shape[1])
                
                patch = patch * grad_patch
                
                result[real_h0:real_h_end, real_w0:real_w_end] += patch[ph_0:ph_end,pw_0:pw_end,...]
                sum_result[real_h0:real_h_end, real_w0:real_w_end] += (np.ones_like(patch) * grad_patch)[ph_0:ph_end,pw_0:pw_end]

                
                
        sum_result[sum_result == 0] = 1
        # correct weighting
        result = result / sum_result
        return result

    
    def createModelOutputGradient(self):
        # creates an 2d-array with 1s in the center and decaying to 0 at the borders
        width = self.parent.augmentation.outputwidth
        height = self.parent.augmentation.outputheight
        cutoff = self.parent.ModelOverlap//2
        
        a = np.arange(width)
        b = np.minimum(a,a[::-1])
        c = np.sqrt(b.reshape(1,width)*b.reshape(width,1))
        outarray = c / c[width//2,cutoff]
        outarray[outarray>1] = 1
        outarray = cv2.resize(outarray, (width,height))
        outarray = outarray.reshape(width,height,1)
        
        return outarray
    
    def Predict(self, image):
        if len(image.shape) == 2:
            image = image[...,np.newaxis]

        image = self.parent.data.preprocessImage(image)
        image = image[np.newaxis, ...]
        
        # convert_to_tensor is not necessary but improves performance because it avoids retracing
        pred = self.parent.Model.predict(tf.convert_to_tensor(self.parent.augmentation.checkModelSizeCompatibility(image),np.float32))
        pred = self.convert2Image(pred)

        pred = pred.astype('uint8')
        return pred

    def getImageSize4ModelInput(self, inputWidth, inputHeight):
        width = int(inputWidth*self.parent.ImageScaleFactor)
        height= int(inputHeight*self.parent.ImageScaleFactor)
        return width, height

    def setBackbone(self, backbone):
        self.backbone = backbone

    def setArchitecture(self, architecture):
        self.architecture = architecture

    def getBackbones(self):
        if self.architecture == 'Deeplabv3+':
            return deeplab_backbones()
        else:
            return Backbones.models_names()

    def getArchitectures(self):
        return ['UNet','FPN', 'Linknet', 'PSPNet', 'Deeplabv3+']

    def getPixelModel(self, nclasses, channels, activation):
        addextraInputLayer = False
        if self.pretrained:
            if self.architecture == 'Deeplabv3+':
                weights = 'pascal_voc'
                self.preprocessingfnc = deeplab_preprocess
            else:
                weights = 'imagenet'
                self.preprocessingfnc = get_preprocessing(self.backbone)
            inputsize = (None, None, 3)
            if channels != 3:
                addextraInputLayer = True 
        else:
            weights = None
            self.preprocessingfnc = None
            inputsize = (None, None, channels)


        if self.architecture == 'UNet':
            model = Unet(backbone_name=self.backbone, input_shape=inputsize, classes = nclasses, encoder_weights=weights, activation=activation)
        elif self.architecture == 'FPN':
            model = FPN(backbone_name=self.backbone, input_shape=inputsize, classes = nclasses, encoder_weights=weights, activation=activation)
        elif self.architecture == 'Linknet':
            model = Linknet(backbone_name=self.backbone, input_shape=inputsize, classes = nclasses, encoder_weights=weights, activation=activation)
        elif self.architecture == 'PSPNet':
            # must be divisible by 48
            width = self.parent.augmentation.outputwidth - self.parent.augmentation.outputwidth % 48
            height = self.parent.augmentation.outputheight - self.parent.augmentation.outputheight % 48
            # self.parent.augmentation.outputwidth = width
            # self.parent.augmentation.outputheight = height
            if self.pretrained:
                inputsize = (width, height, 3)
            else:
                inputsize = (width, height, channels)
            model = PSPNet(backbone_name = self.backbone, input_shape=inputsize, classes = nclasses, encoder_weights=weights, activation=activation)
        elif self.architecture == 'Deeplabv3+':
            if self.pretrained:
                # self.parent.augmentation.outputwidth = 512
                # self.parent.augmentation.outputheight = 512
                inputsize = (512, 512, 3)
            else:
                inputsize = (self.parent.augmentation.outputwidth, self.parent.augmentation.outputheight, channels)
            model = Deeplabv3(weights=weights, input_shape=inputsize, classes=nclasses, backbone=self.backbone, OS=16, alpha=1., activation=activation)

        else:
            raise ValueError("not a supported network architecture")

        if addextraInputLayer:
            input_layer = tf.keras.layers.Input(shape=(None, None, channels))
            layer0 = tf.keras.layers.Conv2D(3, (1, 1))(input_layer) 
            out = model(layer0)
            model = tf.keras.models.Model(input_layer, out, name=model.name)
            
        self.setModelInputSize()
        return model
    
    def setModelInputSize(self):
        if self.architecture == 'Deeplabv3+':
            if self.pretrained:
                self.parent.augmentation.outputwidth = 512
                self.parent.augmentation.outputheight = 512
                
        elif self.architecture == 'PSPNet':
            # must be divisible by 48
            self.parent.augmentation.outputwidth = self.parent.augmentation.outputwidth - self.parent.augmentation.outputwidth % 48
            self.parent.augmentation.outputheight = self.parent.augmentation.outputheight - self.parent.augmentation.outputheight % 48
                
                
    
    def loadmodel(self, path):
        if self.pretrained:
            if self.architecture == 'Deeplabv3+':
                self.preprocessingfnc = deeplab_preprocess
            else:
                self.preprocessingfnc = get_preprocessing(self.backbone)
        else:
            self.preprocessingfnc = None
        self.setModelInputSize()
        return load_model(path)
    
    @abstractmethod
    def convert2Image(self,prediction):
        pass
    
    @abstractmethod
    def resizeLabel(self, label, shape):
        pass
        
    @abstractmethod
    def LabelDistance(self,l1, l2):
        pass

    @abstractmethod
    def LoadShapes(self,filename):
        pass


# -*- coding: utf-8 -*-
"""
Created on Thu Oct  8 16:32:20 2020

@author: Koerber
"""

from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *
from ui.style import styleForm
from ui.ui_utils import LabelledSpinBox

class Window(object):
    def setupUi(self, Form):

        Form.setWindowTitle('Segmentation Settings') 
        styleForm(Form)  

        self.centralWidget = QWidget(Form)
        
        vlayout = QVBoxLayout(self.centralWidget)
        vlayout.setContentsMargins(3, 3, 3, 3)
        vlayout.setSpacing(2)
        
        self.CBDitanceMap = QCheckBox("Separate contours",self.centralWidget)
        self.CBDitanceMap.setObjectName('nn_CreateWeightedDistanceMap')
        self.CBDitanceMap.setToolTip('Check to separate adjacent contours')
        
        self.SBWeight = LabelledSpinBox ('Inter countour weighting            ', self.centralWidget)
        self.SBWeight.setObjectName('nn_SeparationWeight_seg')
        self.SBWeight.SpinBox.setRange(2,50)
        
        self.SBDist = LabelledSpinBox ('Inter contour distance               ', self.centralWidget)
        self.SBDist.setObjectName('nn_SeparationDist_seg')
        self.SBDist.SpinBox.setRange(1,50)
        
        vlayout.addWidget(self.CBDitanceMap)
        vlayout.addWidget(self.SBWeight)
        vlayout.addWidget(self.SBDist)

        self.CBIgnoreBackgroundTiles = QCheckBox("Prefer labelled parts",self.centralWidget)
        self.CBIgnoreBackgroundTiles.setToolTip('When checked image parts which contain only background are ignored during training')
        self.CBIgnoreBackgroundTiles.setObjectName('nn_IgnoreBackground_seg')

        self.SBDiscard = LabelledSpinBox ('Discard up to x background tiles', self.centralWidget)
        self.SBDiscard.setObjectName('nn_DiscardBGTiles_seg')
        self.SBDiscard.SpinBox.setRange(1,20)

        self.SBMinPixels = LabelledSpinBox ('Minimum required labelled pixels', self.centralWidget)
        self.SBMinPixels.setObjectName('nn_MinLabelledPixels_seg')
        self.SBMinPixels.SpinBox.setRange(1,1000)


        vlayout.addWidget(self.CBIgnoreBackgroundTiles)
        vlayout.addWidget(self.SBDiscard)
        vlayout.addWidget(self.SBMinPixels)

        self.centralWidget.setLayout(vlayout)

        self.centralWidget.setFixedSize(vlayout.sizeHint())
        Form.setFixedSize(vlayout.sizeHint())
        


class SegmentationSettingsWindow(QMainWindow, Window):
    def __init__(self, parent ):
        super(SegmentationSettingsWindow, self).__init__(parent)
        self.parent = parent
        self.setupUi(self)

        self.CBIgnoreBackgroundTiles.stateChanged.connect(self.IgnoreBackgroundChanged)
        self.SBDiscard.SpinBox.valueChanged.connect(self.DiscardChanged)
        self.SBMinPixels.SpinBox.valueChanged.connect(self.MinPixelsChanged)

        self.CBDitanceMap.stateChanged.connect(self.ShapeSeparationChanged)
        self.SBWeight.SpinBox.valueChanged.connect(self.WeightChanged)
        self.SBDist.SpinBox.valueChanged.connect(self.DistChanged)
        
        
        
        self.CBIgnoreBackgroundTiles.setChecked(self.parent.parent.dl.augmentation.seg_ignoreBackground)
        self.CBDitanceMap.setChecked(self.parent.parent.dl.seg_useWeightedDistanceMap)
        
    def show(self):
        self.CBIgnoreBackgroundTiles.setChecked(self.parent.parent.dl.augmentation.seg_ignoreBackground)
        self.SBDiscard.SpinBox.setValue(self.parent.parent.dl.augmentation.seg_ignorecycles)
        self.SBMinPixels.SpinBox.setValue(self.parent.parent.dl.augmentation.seg_minRequiredPixels)
        
        self.SBWeight.SpinBox.setValue(self.parent.parent.dl.seg_border_weight)
        self.SBDist.SpinBox.setValue(self.parent.parent.dl.seg_border_dist)

        self.CBDitanceMap.setChecked(self.parent.parent.dl.seg_useWeightedDistanceMap)
        super(SegmentationSettingsWindow, self).show()
        
    def ShapeSeparationChanged(self):
        self.parent.parent.dl.seg_useWeightedDistanceMap = self.CBDitanceMap.isChecked()
        self.SBWeight.setEnabled(self.parent.parent.dl.seg_useWeightedDistanceMap)
        self.SBDist.setEnabled(self.parent.parent.dl.seg_useWeightedDistanceMap)
        
        
    def WeightChanged(self):
        self.parent.parent.dl.seg_border_weight = self.SBWeight.SpinBox.value()
        
    def DistChanged(self):
        self.parent.parent.dl.seg_border_dist = self.SBDist.SpinBox.value()

    def IgnoreBackgroundChanged(self):
        self.parent.parent.dl.augmentation.seg_ignoreBackground = self.CBIgnoreBackgroundTiles.isChecked()
        if self.CBIgnoreBackgroundTiles.isChecked():
            self.SBMinPixels.setEnabled(True)
            self.SBDiscard.setEnabled(True)
        else:
            self.SBMinPixels.setEnabled(False)
            self.SBDiscard.setEnabled(False)

    def MinPixelsChanged(self):
        self.parent.parent.dl.augmentation.seg_minRequiredPixels = self.SBMinPixels.SpinBox.value()

    def DiscardChanged(self):
        self.parent.parent.dl.augmentation.seg_ignorecycles = self.SBDiscard.SpinBox.value()


from dl.models.keras_applications import get_submodules_from_kwargs

from ._common_blocks import Conv2dBn
from ._utils import freeze_model
from ..backbones.backbones_factory import Backbones

backend = None
layers = None
models = None
keras_utils = None


# ---------------------------------------------------------------------
#  Utility functions
# ---------------------------------------------------------------------

def get_submodules():
    return {
        'backend': backend,
        'models': models,
        'layers': layers,
        'utils': keras_utils,
    }


def check_input_shape(input_shape, factor):
    if input_shape is None:
        raise ValueError("Input shape should be a tuple of 3 integers, not None!")

    h, w = input_shape[:2] if backend.image_data_format() == 'channels_last' else input_shape[1:]
    min_size = factor * 6

    is_wrong_shape = (
            h % min_size != 0 or w % min_size != 0 or
            h < min_size or w < min_size
    )

    if is_wrong_shape:
        raise ValueError('Wrong shape {}, input H and W should '.format(input_shape) +
                         'be divisible by `{}`'.format(min_size))


# ---------------------------------------------------------------------
#  Blocks
# ---------------------------------------------------------------------

def Conv1x1BnReLU(filters, use_batchnorm, name=None):
    kwargs = get_submodules()

    def wrapper(input_tensor):
        return Conv2dBn(
            filters,
            kernel_size=1,
            activation='relu',
            kernel_initializer='he_uniform',
            padding='same',
            use_batchnorm=use_batchnorm,
            name=name,
            **kwargs
        )(input_tensor)

    return wrapper


def SpatialContextBlock(
        level,
        conv_filters=512,
        pooling_type='avg',
        use_batchnorm=True,
):
    if pooling_type not in ('max', 'avg'):
        raise ValueError('Unsupported pooling type - `{}`.'.format(pooling_type) +
                         'Use `avg` or `max`.')

    Pooling2D = layers.MaxPool2D if pooling_type == 'max' else layers.AveragePooling2D

    pooling_name = 'psp_level{}_pooling'.format(level)
    conv_block_name = 'psp_level{}'.format(level)
    upsampling_name = 'psp_level{}_upsampling'.format(level)

    def wrapper(input_tensor):
        # extract input feature maps size (h, and w dimensions)
        input_shape = backend.int_shape(input_tensor)
        spatial_size = input_shape[1:3] if backend.image_data_format() == 'channels_last' else input_shape[2:]

        # Compute the kernel and stride sizes according to how large the final feature map will be
        # When the kernel factor and strides are equal, then we can compute the final feature map factor
        # by simply dividing the current factor by the kernel or stride factor
        # The final feature map sizes are 1x1, 2x2, 3x3, and 6x6.
        pool_size = up_size = [spatial_size[0] // level, spatial_size[1] // level]

        x = Pooling2D(pool_size, strides=pool_size, padding='same', name=pooling_name)(input_tensor)
        x = Conv1x1BnReLU(conv_filters, use_batchnorm, name=conv_block_name)(x)
        x = layers.UpSampling2D(up_size, interpolation='bilinear', name=upsampling_name)(x)
        return x

    return wrapper


# ---------------------------------------------------------------------
#  PSP Decoder
# ---------------------------------------------------------------------

def build_psp(
        backbone,
        psp_layer_idx,
        pooling_type='avg',
        conv_filters=512,
        use_batchnorm=True,
        final_upsampling_factor=8,
        classes=21,
        activation='softmax',
        dropout=None,
):
    input_ = backbone.input
    x = (backbone.get_layer(name=psp_layer_idx).output if isinstance(psp_layer_idx, str)
         else backbone.get_layer(index=psp_layer_idx).output)

    # build spatial pyramid
    x1 = SpatialContextBlock(1, conv_filters, pooling_type, use_batchnorm)(x)
    x2 = SpatialContextBlock(2, conv_filters, pooling_type, use_batchnorm)(x)
    x3 = SpatialContextBlock(3, conv_filters, pooling_type, use_batchnorm)(x)
    x6 = SpatialContextBlock(6, conv_filters, pooling_type, use_batchnorm)(x)

    # aggregate spatial pyramid
    concat_axis = 3 if backend.image_data_format() == 'channels_last' else 1
    x = layers.Concatenate(axis=concat_axis, name='psp_concat')([x, x1, x2, x3, x6])
    x = Conv1x1BnReLU(conv_filters, use_batchnorm, name='aggregation')(x)

    # model regularization
    if dropout is not None:
        x = layers.SpatialDropout2D(dropout, name='spatial_dropout')(x)

    # model head
    x = layers.Conv2D(
        filters=classes,
        kernel_size=(3, 3),
        padding='same',
        kernel_initializer='glorot_uniform',
        name='final_conv',
    )(x)

    x = layers.UpSampling2D(final_upsampling_factor, name='final_upsampling', interpolation='bilinear')(x)
    x = layers.Activation(activation, name=activation)(x)

    model = models.Model(input_, x)

    return model


# ---------------------------------------------------------------------
#  PSP Model
# ---------------------------------------------------------------------

def PSPNet(
        backbone_name='vgg16',
        input_shape=(384, 384, 3),
        classes=21,
        activation='softmax',
        weights=None,
        encoder_weights='imagenet',
        encoder_freeze=False,
        downsample_factor=8,
        psp_conv_filters=512,
        psp_pooling_type='avg',
        psp_use_batchnorm=True,
        psp_dropout=None,
        **kwargs
):
    """PSPNet_ is a fully convolution neural network for image semantic segmentation

    Args:
        backbone_name: name of classification model used as feature
                extractor to build segmentation model.
        input_shape: shape of input data/image ``(H, W, C)``.
            ``H`` and ``W`` should be divisible by ``6 * downsample_factor`` and **NOT** ``None``!
        classes: a number of classes for output (output shape - ``(h, w, classes)``).
        activation: name of one of ``keras.activations`` for last model layer
                (e.g. ``sigmoid``, ``softmax``, ``linear``).
        weights: optional, path to model weights.
        encoder_weights: one of ``None`` (random initialization), ``imagenet`` (pre-training on ImageNet).
        encoder_freeze: if ``True`` set all layers of encoder (backbone model) as non-trainable.
        downsample_factor: one of 4, 8 and 16. Downsampling rate or in other words backbone depth
            to construct PSP module on it.
        psp_conv_filters: number of filters in ``Conv2D`` layer in each PSP block.
        psp_pooling_type: one of 'avg', 'max'. PSP block pooling type (maximum or average).
        psp_use_batchnorm: if ``True``, ``BatchNormalisation`` layer between ``Conv2D`` and ``Activation`` layers
                is used.
        psp_dropout: dropout rate between 0 and 1.

    Returns:
        ``keras.models.Model``: **PSPNet**

    .. _PSPNet:
        https://arxiv.org/pdf/1612.01105.pdf

    """

    global backend, layers, models, keras_utils
    backend, layers, models, keras_utils = get_submodules_from_kwargs(kwargs)

    # control image input shape
    check_input_shape(input_shape, downsample_factor)

    backbone = Backbones.get_backbone(
        backbone_name,
        input_shape=input_shape,
        weights=encoder_weights,
        include_top=False,
        **kwargs
    )

    feature_layers = Backbones.get_feature_layers(backbone_name, n=3)

    if downsample_factor == 16:
        psp_layer_idx = feature_layers[0]
    elif downsample_factor == 8:
        psp_layer_idx = feature_layers[1]
    elif downsample_factor == 4:
        psp_layer_idx = feature_layers[2]
    else:
        raise ValueError('Unsupported factor - `{}`, Use 4, 8 or 16.'.format(downsample_factor))

    model = build_psp(
        backbone,
        psp_layer_idx,
        pooling_type=psp_pooling_type,
        conv_filters=psp_conv_filters,
        use_batchnorm=psp_use_batchnorm,
        final_upsampling_factor=downsample_factor,
        classes=classes,
        activation=activation,
        dropout=psp_dropout,
    )

    # lock encoder weights for fine-tuning
    if encoder_freeze:
        freeze_model(backbone, **kwargs)

    # loading model weights
    if weights is not None:
        model.load_weights(weights)

    return model



# -*- coding: utf-8 -*-
"""
Created on Wed Jun 10 16:37:31 2020

@author: Koerber
"""

from PyQt5.QtCore import *
from abc import ABC, abstractmethod
import traceback
import sys


class Observer(ABC):

    @abstractmethod
    def Error(self, exctype, value, traceback):
        pass
    @abstractmethod
    def Finished(self):
        pass
    @abstractmethod
    def Result(self,result):
        pass
    @abstractmethod
    def Progress(self,i):
        pass
    @abstractmethod
    def Started(self):
        pass



class InterfaceSignals(QObject):
    finished = pyqtSignal()
    error = pyqtSignal(tuple)
    result = pyqtSignal(object)
    progress = pyqtSignal()
    start = pyqtSignal()
    epoch_end = pyqtSignal(int)

class dlObservable():
    def __init__(self):
        self.observers = []
        
    def attachObserver(self,o):
        self.observers.append(o)
        
    def detachObserver(self,o):
        self.observers.remove(o)
        
    def notifyError(self):
        for o in self.observers:
            exctype, value, tb = sys.exc_info()
            o.Error(exctype, value, tb)
            
    def notifyPredictionFinished(self, prediction):
        for o in self.observers:
            o.Result(prediction)            

    def notifyTrainingStarted(self):
        for o in self.observers:
            o.Started()            

    def notifyTrainingFinished(self):
        for o in self.observers:
            o.Finished()
        
    def notifyTrainingResult(self, result):
        for o in self.observers:
            o.Result(result)
        
    def notifyTrainingProgress(self):
        for o in self.observers:
            o.Progress()
            
    def notifyEpochEnd(self, epoch):
        for o in self.observers:
            o.EpochEnd(epoch)
    

class QtObserver(Observer):
    def __init__(self):
        self.signals = InterfaceSignals()  
        
    def Error(self,exctype, value, traceback):
        self.signals.error.emit((exctype, value, traceback))
        
    def Finished(self):
        self.signals.finished.emit()
        
    def Result(self, result):
        self.signals.result.emit(result)
        
    def Progress(self):
        self.signals.progress.emit()
        
    def Started(self):
        self.signals.start.emit()
        
    def EpochEnd(self, epoch):
        self.signals.epoch_end.emit(epoch)
        
    
        


# -*- coding: utf-8 -*-
"""
Created on Wed Jul 22 16:15:28 2020

@author: Koerber
"""

from dl.method.mode import LearningMode, dlMode
from dl.method.pixelbasedprediction import PixelBasedPrediction
from skimage.feature import peak_local_max
import utils.shapes.Point as Point
from utils.shapes.Contour import removeBackground
import numpy as np
import cv2
from dl.loss.regression_losses import RegressionLosses
from dl.metric.regression_metrics import RegressionMetrics



class ObjectCounting(PixelBasedPrediction, LearningMode):
    def __init__(self, parent):
        PixelBasedPrediction.__init__(self,parent)  
        LearningMode.__init__(self, parent)
        self.type = dlMode.Object_Counting
        self.metric = RegressionMetrics(parent)
        self.loss = RegressionLosses(parent)

        self.kernel = None
        self.scalefactor = 1

        self.maxDetectionLimit = 5000

        self.architecture = 'UNet'
        self.backbone = 'resnet50'
        
    def LoadLabel(self, filename, height, width, rmBackground = False):
        label = np.zeros((height, width, 1), np.uint8)
        points, bg = Point.loadPoints(filename)
        label = Point.drawPointsToLabel(label, points, bg)
        if rmBackground:
            return removeBackground(label)
        return label
    
    def setObjectSize(self):
        # stdev should be between 1 and 10 and equals the expansion of an object
        # peakval is the max value of a single object at its coordinates
        k = cv2.getGaussianKernel(67, self.parent.od_kernel_stdev) 
        self.kernel = k*np.transpose(k)
        self.scalefactor = self.parent.od_peak_val/np.max(self.kernel)
        
    def prepreprocessLabel(self, label):
        # should we create heatmap here?
        return label
        
    def preprocessLabel(self, label):
        label = label.astype('float')
        self.setObjectSize()
        if self.parent.NumClasses > 2:
            # convert to an 1-hot encoded heatmap representing with the point coordinates in its maxima
            one_hot = np.eye(self.parent.NumClasses)[np.squeeze(label.astype(int), axis=-1)]
            for i in range(label.shape[0]):
                one_hot[i,...] = cv2.filter2D(one_hot[i,...],-1, self.kernel)
                one_hot[i,...,0] = one_hot[i,...,0]-(1-np.max(self.kernel))
                label = one_hot
        else:
            for i in range(label.shape[0]):
                label[i,...,0] = cv2.filter2D(label[i,...],-1, self.kernel)
            
        return label*self.scalefactor

    def getModel(self, nclasses, channels):
        return PixelBasedPrediction.getPixelModel(self, nclasses, channels, None)

 
    def extractShapesFromPrediction(self, prediction, unused, offset=(0,0)):
        return Point.extractPointsFromLabel(prediction, offset)
    
    def saveShapes(self, contours, path):
        Point.savePoints(contours, path)
        
    def convert2Image(self,prediction):
        prediction = np.squeeze(prediction)
        # threshold?
        thresh = 1
        calcResult = True
        # we need to add emergency break here, otherwise too many objects might be detected and break program
        while calcResult and thresh < 100:
            result = np.zeros(prediction.shape[0:2])
            prediction[prediction < thresh] = 0

            if self.parent.NumClasses > 2:
                maxarg = np.argmax(prediction, axis = 2)
                for i in range(1,prediction.shape[2]):
                    peaks = peak_local_max(prediction[...,i], min_distance=3, indices = False)
                    result[(maxarg==i) & peaks] = i 
            else:
                peaks = peak_local_max(prediction, min_distance=3, indices = False)
                result[peaks] = 1
            calcResult = np.count_nonzero(result) > self.maxDetectionLimit
            thresh += 1

        return result
    
    def resizeLabel(self, label, shape):
        # every pixel != 0 will create exactly one pixel != 0 in the rezised image at the corresponding position
        newlabel = np.zeros((shape[1],shape[0]), dtype=label.dtype)
        if self.parent.NumClasses > 2:
            for i in range(1,np.max(label)+1):
                indices = np.where(label==i)
                idx_0 = (indices[0] / label.shape[0] * shape[1]).astype(int)
                idx_1 = (indices[1] / label.shape[1] * shape[0]).astype(int)
                newlabel[(idx_0, idx_1)] = i
        else:
            indices = np.where(label==1) 
            idx_0 = (indices[0] / label.shape[0] * shape[1]).astype(int)
            idx_1 = (indices[1] / label.shape[1] * shape[0]).astype(int)

            newlabel[(idx_0, idx_1)] = 1
        return newlabel
    
    def LabelDistance(self,l1, l2):
        return Point.distance_p1_p2(l1.coordinates,l2.coordinates)
    
    def LoadShapes(self,filename):
        points, bg = Point.loadPoints(filename)
        return points
        
    def countLabelWeight(self, label):
        classValues = np.zeros((self.parent.NumClasses_real), dtype = np.float)
        for i in range(1,self.parent.NumClasses_real):
            classValues[i] = self.parent.od_kernel_stdev*np.count_nonzero((label == i))
        classValues[0] = label.size - self.parent.od_kernel_stdev*np.count_nonzero(label)
        return classValues


# -*- coding: utf-8 -*-
"""
Created on Wed Feb 19 13:43:00 2020

@author: Koerber
"""


from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *
from ui.ui_utils import LabelledSpinBox, LabelledDoubleSpinBox
from ui.style import styleForm



class Window(object):
    def setupUi(self, Form):

        Form.setWindowTitle('Augmentation') 
        styleForm(Form)     

        self.centralWidget = QWidget(Form)


        self.vlayout = QVBoxLayout(self.centralWidget)
        self.vlayout.setContentsMargins(3, 3, 3, 3)
        self.vlayout.setSpacing(0)
        self.centralWidget.setLayout(self.vlayout)

        self.augmentations = self.AugmentationGroup()
        self.augmentations.setCheckable(True)
        self.augmentations.setObjectName('nn_Augmentations')
        self.vlayout.addWidget(self.augmentations)

        self.centralWidget.setFixedSize(self.vlayout.sizeHint())
        Form.setFixedSize(self.vlayout.sizeHint())


    def AugmentationGroup(self):
       
        groupBox = QGroupBox("Augmentations")

        layout = QVBoxLayout(self.centralWidget)
        layout.setSpacing(1)
        layout.setContentsMargins(1, 1, 1, 1)
        sizelayout = QHBoxLayout(self.centralWidget)
        sizelayout.setSpacing(0)
        sizelayout.setContentsMargins(1, 1, 1, 1)
        self.SBModelInputSize_x = LabelledSpinBox(' Input width', self.centralWidget)
        self.SBModelInputSize_x.setToolTip('Set model input width')
        self.SBModelInputSize_x.setObjectName('nn_ModelInput_X')
        self.SBModelInputSize_y = LabelledSpinBox(' Input height', self.centralWidget)
        self.SBModelInputSize_y.setToolTip('Set model input height')
        self.SBModelInputSize_y.setObjectName('nn_ModelInput_Y')

        sizelayout.addWidget(self.SBModelInputSize_x)
        sizelayout.addWidget(self.SBModelInputSize_y)
        layout.addItem(sizelayout)

        hlayout = QHBoxLayout(self.centralWidget)
        self.Flip = self.FlipGroup()
        hlayout.addWidget(self.Flip)
        self.Noise = self.NoiseGroup()
        hlayout.addWidget(self.Noise)
        layout.addItem(hlayout)

        self.Affine = self.AffineGroup()
        self.Affine.setObjectName('nn_AffineTransformation')
        self.Affine.setCheckable(True)
        layout.addWidget(self.Affine)

        groupBox.setLayout(layout)

        return groupBox


    def FlipGroup(self):
        groupBox = QGroupBox("Flip")
        layout = QVBoxLayout(self.centralWidget)
        layout.setSpacing(1)
        layout.setContentsMargins(0, 0, 0, 0)
        self.CBFliplr = QCheckBox("Flip horizontal",self.centralWidget)
        self.CBFliplr.setObjectName('nn_Flip_horizontal')
        self.CBFliplr.setToolTip('Check to flip images horizontally')
        self.CBFliptb = QCheckBox("Flip vertical",self.centralWidget)
        self.CBFliptb.setToolTip('Check to flip images vertically')
        self.CBFliptb.setObjectName('nn_Flip_vertical')

        layout.addWidget(self.CBFliplr)
        layout.addWidget(self.CBFliptb)


        groupBox.setLayout(layout)
        return groupBox


    def NoiseGroup(self):
        groupBox = QGroupBox("Noise")
        layout = QVBoxLayout(self.centralWidget)
        layout.setSpacing(1)
        layout.setContentsMargins(0, 0, 0, 0)
        self.SBNoise = LabelledSpinBox(' Blur (%)', self.centralWidget)
        self.SBNoise.SpinBox.setRange(0,100)
        self.SBNoise.setObjectName('nn_Blurring')
        self.SBNoise.setToolTip('Percentage to apply gaussian blur to the input image')

        self.SBPieceWise = LabelledSpinBox(' Piecewise (%)', self.centralWidget)
        self.SBPieceWise.SpinBox.setRange(0,100)
        self.SBPieceWise.setObjectName('nn_Piecewise_Affine')
        self.SBPieceWise.setToolTip('Percentage to apply piecewise affine transformation')

        self.SBDropout = LabelledSpinBox(' Dropout (%)', self.centralWidget)
        self.SBDropout.SpinBox.setRange(0,100)
        self.SBDropout.setObjectName('nn_ImageDropout')
        self.SBDropout.setToolTip('Percentage to apply piecewise dropout')

        layout.addWidget(self.SBNoise)
        layout.addWidget(self.SBPieceWise)
        layout.addWidget(self.SBDropout)

        groupBox.setLayout(layout)
        return groupBox

    def AffineGroup(self):

        groupBox = QGroupBox("Affine")
        layout = QVBoxLayout(self.centralWidget)
        layout.setSpacing(1)
        layout.setContentsMargins(1, 1, 1, 1)
        hlayout = QHBoxLayout(self.centralWidget)
        self.SBPercentage = LabelledSpinBox(' Apply Affine (%)', self.centralWidget)
        self.SBPercentage.setToolTip('Percentage to apply affine transformation')
        self.SBPercentage.SpinBox.setRange(0,100)
        hlayout.addWidget(self.SBPercentage)

        layout.addLayout(hlayout)

        grid = QGridLayout(self.centralWidget)
        grid.addWidget(self.ScaleGroup(), 0,0)
        grid.addWidget(self.TranslateGroup(), 0,1)
        grid.addWidget(self.ShearGroup(), 1,0)
        grid.addWidget(self.RotateGroup(), 1,1)
        layout.addItem(grid)
        groupBox.setLayout(layout)
        return groupBox

    def ScaleGroup(self):
        groupBox = QGroupBox("Scale")
        layout = QVBoxLayout(self.centralWidget)
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        self.SBScaleMin = LabelledDoubleSpinBox(' Scale Min (factor)', self.centralWidget)
        self.SBScaleMin.setToolTip('Set scale maximum percentage')
        self.SBScaleMin.setObjectName('nn_Scale_min')
        self.SBScaleMin.SpinBox.setSingleStep(0.1)
        self.SBScaleMin.SpinBox.setDecimals(1)
        self.SBScaleMin.SpinBox.setRange(0,1)

        self.SBScaleMax = LabelledDoubleSpinBox(' Scale Max (factor)', self.centralWidget)
        self.SBScaleMax.setToolTip('Set scale minimum percentage')
        self.SBScaleMax.setObjectName('nn_Scale_max')
        self.SBScaleMax.SpinBox.setSingleStep(0.1)
        self.SBScaleMax.SpinBox.setDecimals(1)
        self.SBScaleMax.SpinBox.setRange(1,10)

        layout.addWidget(self.SBScaleMin)
        layout.addWidget(self.SBScaleMax)


        groupBox.setLayout(layout)
        return groupBox

    def TranslateGroup(self):
        groupBox = QGroupBox("Translate")
        layout = QVBoxLayout(self.centralWidget)
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        self.SBTranslateMin = LabelledDoubleSpinBox(' Translate Min (%)', self.centralWidget)
        self.SBTranslateMin.setToolTip('Set translation minimun percent')
        self.SBTranslateMin.setObjectName('nn_Translate_min')
        self.SBTranslateMin.SpinBox.setSingleStep(0.1)
        self.SBTranslateMin.SpinBox.setDecimals(1)
        self.SBTranslateMin.SpinBox.setRange(-1,0)

        self.SBTranslateMax = LabelledDoubleSpinBox(' Translate Max (%)', self.centralWidget)
        self.SBTranslateMax.setToolTip('Set translation minimun percent')
        self.SBTranslateMax.SpinBox.setSingleStep(0.1)
        self.SBTranslateMax.setObjectName('nn_Translate_max')
        self.SBTranslateMax.SpinBox.setDecimals(1)
        self.SBTranslateMax.SpinBox.setRange(0,1)

        layout.addWidget(self.SBTranslateMin)
        layout.addWidget(self.SBTranslateMax)

        groupBox.setLayout(layout)
        return groupBox

    def ShearGroup(self):
        groupBox = QGroupBox("Shear")
        layout = QVBoxLayout(self.centralWidget)
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        self.SBShearMin = LabelledSpinBox(' Shear Min (deg)', self.centralWidget)
        self.SBShearMin.setToolTip('Set shear minimum')
        self.SBShearMin.setObjectName('nn_Shear_min')
        self.SBShearMin.SpinBox.setRange(-45,0)

        self.SBShearMax = LabelledSpinBox(' Shear Max (deg)', self.centralWidget)
        self.SBShearMax.setToolTip('Set shear maximum')
        self.SBShearMax.setObjectName('nn_Shear_max')
        self.SBShearMax.SpinBox.setRange(0,45)

        layout.addWidget(self.SBShearMin)
        layout.addWidget(self.SBShearMax)

        groupBox.setLayout(layout)
        return groupBox

    def RotateGroup(self):
        groupBox = QGroupBox("Rotate")
        layout = QVBoxLayout(self.centralWidget)
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        self.SBRotMin = LabelledSpinBox(' Rotation Min (deg)', self.centralWidget)
        self.SBRotMin.setToolTip('Set rotation range minimum')
        self.SBRotMin.setObjectName('nn_Rotate_min')
        self.SBRotMin.SpinBox.setRange(-45,0)

        self.SBRotMax = LabelledSpinBox(' Rotation Max (deg)', self.centralWidget)
        self.SBRotMax.setToolTip('Set rotation range maximum')
        self.SBRotMax.setObjectName('nn_Rotate_max')
        self.SBRotMax.SpinBox.setRange(0,45)

        layout.addWidget(self.SBRotMin)
        layout.addWidget(self.SBRotMax)

        groupBox.setLayout(layout)
        return groupBox





class AugmentWindow(QMainWindow, Window):
    def __init__(self, parent ):
        super(AugmentWindow, self).__init__(parent)
        self.parent = parent
        self.setupUi(self)
        
        
        self.SBModelInputSize_x.SpinBox.setRange(self.parent.dl.split_factor,1024)
        self.SBModelInputSize_y.SpinBox.setRange(self.parent.dl.split_factor,1024)
        self.SBModelInputSize_x.SpinBox.setSingleStep(self.parent.dl.split_factor)
        self.SBModelInputSize_y.SpinBox.setSingleStep(self.parent.dl.split_factor)
        self.SBModelInputSize_x.SpinBox.setValue(self.parent.dl.augmentation.outputwidth)
        self.SBModelInputSize_y.SpinBox.setValue(self.parent.dl.augmentation.outputheight)
        
        
        self.CBFliplr.setChecked(self.parent.dl.augmentation.flip_horz)
        self.CBFliptb.setChecked(self.parent.dl.augmentation.flip_vert)
        
        self.SBNoise.SpinBox.setValue(self.parent.dl.augmentation.prob_gaussian *100)
        self.SBPieceWise.SpinBox.setValue(self.parent.dl.augmentation.prob_piecewise *100)
        self.SBDropout.SpinBox.setValue(self.parent.dl.augmentation.prob_dropout *100)
        
        self.SBPercentage.SpinBox.setValue(self.parent.dl.augmentation.prob_affine *100)
        self.SBScaleMin.SpinBox.setValue(self.parent.dl.augmentation.scale_min)
        self.SBScaleMax.SpinBox.setValue(self.parent.dl.augmentation.scale_max)
        self.SBTranslateMin.SpinBox.setValue(self.parent.dl.augmentation.translate_min)
        self.SBTranslateMax.SpinBox.setValue(self.parent.dl.augmentation.translate_max)
        self.SBShearMin.SpinBox.setValue(self.parent.dl.augmentation.shear_min)
        self.SBShearMax.SpinBox.setValue(self.parent.dl.augmentation.shear_max)
        self.SBRotMin.SpinBox.setValue(self.parent.dl.augmentation.rotate_min)
        self.SBRotMax.SpinBox.setValue(self.parent.dl.augmentation.rotate_max)
        
        
        self.SBModelInputSize_x.SpinBox.valueChanged.connect(self.size)
        self.SBModelInputSize_y.SpinBox.valueChanged.connect(self.size)
        
        self.CBFliplr.stateChanged.connect(self.flip)
        self.CBFliptb.stateChanged.connect(self.flip)
        
        self.SBNoise.SpinBox.valueChanged.connect(self.noise)
        self.SBPieceWise.SpinBox.valueChanged.connect(self.piecewise)
        self.SBDropout.SpinBox.valueChanged.connect(self.dropout)
        
        self.SBPercentage.SpinBox.valueChanged.connect(self.affinepercentage)
        self.SBScaleMin.SpinBox.valueChanged.connect(self.scale)
        self.SBScaleMax.SpinBox.valueChanged.connect(self.scale)
        self.SBTranslateMin.SpinBox.valueChanged.connect(self.translate)
        self.SBTranslateMax.SpinBox.valueChanged.connect(self.translate)
        self.SBShearMin.SpinBox.valueChanged.connect(self.shear)
        self.SBShearMax.SpinBox.valueChanged.connect(self.shear)
        self.SBRotMin.SpinBox.valueChanged.connect(self.rotate)
        self.SBRotMax.SpinBox.valueChanged.connect(self.rotate)
        
        
        self.augmentations.setChecked(self.parent.dl.augmentation.enableAll)
        self.augmentations.clicked.connect(self.enableAll)
        
        self.Affine.setChecked(self.parent.dl.augmentation.enableAffine)
        self.Affine.clicked.connect(self.enableAffine)
    
        
    def enableAll(self):
        self.parent.dl.augmentation.enableAll = self.augmentations.isChecked()
    
    def size(self):
        self.parent.dl.augmentation.setoutputwidth(self.SBModelInputSize_x.SpinBox.value())
        self.parent.dl.augmentation.setoutputheight(self.SBModelInputSize_y.SpinBox.value())

        
    def flip(self):
        self.parent.dl.augmentation.flip_horz = self.CBFliplr.isChecked()
        self.parent.dl.augmentation.flip_vert = self.CBFliptb.isChecked()
        
    def noise(self):
        self.parent.dl.augmentation.prob_gaussian = self.SBNoise.SpinBox.value() /100

    def piecewise(self):
        self.parent.dl.augmentation.prob_piecewise = self.SBPieceWise.SpinBox.value() /100
        
    def dropout(self):
        self.parent.dl.augmentation.prob_dropout = self.SBDropout.SpinBox.value() /100
        
    def enableAffine(self):
        self.parent.dl.augmentation.enableAffine = self.Affine.isChecked()
        
    def affinepercentage(self):
        self.parent.dl.augmentation.prob_affine = self.SBPercentage.SpinBox.value() /100
        
    def scale(self):
        self.parent.dl.augmentation.scale_min = self.SBScaleMin.SpinBox.value()
        self.parent.dl.augmentation.scale_max = self.SBScaleMax.SpinBox.value()

    def translate(self):
        self.parent.dl.augmentation.translate_min = self.SBTranslateMin.SpinBox.value()
        self.parent.dl.augmentation.translate_max = self.SBTranslateMax.SpinBox.value()
        
    def shear(self):
        self.parent.dl.augmentation.shear_min = self.SBShearMin.SpinBox.value()
        self.parent.dl.augmentation.shear_max = self.SBShearMax.SpinBox.value()
    
    def rotate(self):
        self.parent.dl.augmentation.rotate_min = self.SBRotMin.SpinBox.value()
        self.parent.dl.augmentation.rotate_max = self.SBRotMax.SpinBox.value()


SMOOTH = 1e-5


# ----------------------------------------------------------------
#   Helpers
# ----------------------------------------------------------------

def _gather_channels(x, indexes, **kwargs):
    """Slice tensor along channels axis by given indexes"""
    backend = kwargs['backend']
    if backend.image_data_format() == 'channels_last':
        x = backend.permute_dimensions(x, (3, 0, 1, 2))
        x = backend.gather(x, indexes)
        x = backend.permute_dimensions(x, (1, 2, 3, 0))
    else:
        x = backend.permute_dimensions(x, (1, 0, 2, 3))
        x = backend.gather(x, indexes)
        x = backend.permute_dimensions(x, (1, 0, 2, 3))
    return x


def get_reduce_axes(per_image, **kwargs):
    backend = kwargs['backend']
    axes = [1, 2] if backend.image_data_format() == 'channels_last' else [2, 3]
    if not per_image:
        axes.insert(0, 0)
    return axes


def gather_channels(*xs, indexes=None, **kwargs):
    """Slice tensors along channels axis by given indexes"""
    if indexes is None:
        return xs
    elif isinstance(indexes, (int)):
        indexes = [indexes]
    xs = [_gather_channels(x, indexes=indexes, **kwargs) for x in xs]
    return xs


def round_if_needed(x, threshold, **kwargs):
    backend = kwargs['backend']
    if threshold is not None:
        x = backend.greater(x, threshold)
        x = backend.cast(x, backend.floatx())
    return x


def average(x, per_image=False, class_weights=None, **kwargs):
    backend = kwargs['backend']
    if per_image:
        x = backend.mean(x, axis=0)
    if class_weights is not None:
        x = x * class_weights
    return backend.mean(x)


# ----------------------------------------------------------------
#   Metric Functions
# ----------------------------------------------------------------

def iou_score(gt, pr, class_weights=1., class_indexes=None, smooth=SMOOTH, per_image=False, threshold=None, **kwargs):
    r""" The `Jaccard index`_, also known as Intersection over Union and the Jaccard similarity coefficient
    (originally coined coefficient de communauté by Paul Jaccard), is a statistic used for comparing the
    similarity and diversity of sample sets. The Jaccard coefficient measures similarity between finite sample sets,
    and is defined as the size of the intersection divided by the size of the union of the sample sets:

    .. math:: J(A, B) = \frac{A \cap B}{A \cup B}

    Args:
        gt: ground truth 4D keras tensor (B, H, W, C) or (B, C, H, W)
        pr: prediction 4D keras tensor (B, H, W, C) or (B, C, H, W)
        class_weights: 1. or list of class weights, len(weights) = C
        class_indexes: Optional integer or list of integers, classes to consider, if ``None`` all classes are used.
        smooth: value to avoid division by zero
        per_image: if ``True``, metric is calculated as mean over images in batch (B),
            else over whole batch
        threshold: value to round predictions (use ``>`` comparison), if ``None`` prediction will not be round

    Returns:
        IoU/Jaccard score in range [0, 1]

    .. _`Jaccard index`: https://en.wikipedia.org/wiki/Jaccard_index

    """

    backend = kwargs['backend']

    gt, pr = gather_channels(gt, pr, indexes=class_indexes, **kwargs)
    pr = round_if_needed(pr, threshold, **kwargs)
    axes = get_reduce_axes(per_image, **kwargs)

    # score calculation
    intersection = backend.sum(gt * pr, axis=axes)
    union = backend.sum(gt + pr, axis=axes) - intersection

    score = (intersection + smooth) / (union + smooth)
    score = average(score, per_image, class_weights, **kwargs)

    return score


def f_score(gt, pr, beta=1, class_weights=1, class_indexes=None, smooth=SMOOTH, per_image=False, threshold=None,
            **kwargs):
    r"""The F-score (Dice coefficient) can be interpreted as a weighted average of the precision and recall,
    where an F-score reaches its best value at 1 and worst score at 0.
    The relative contribution of ``precision`` and ``recall`` to the F1-score are equal.
    The formula for the F score is:

    .. math:: F_\beta(precision, recall) = (1 + \beta^2) \frac{precision \cdot recall}
        {\beta^2 \cdot precision + recall}

    The formula in terms of *Type I* and *Type II* errors:

    .. math:: F_\beta(A, B) = \frac{(1 + \beta^2) TP} {(1 + \beta^2) TP + \beta^2 FN + FP}


    where:
        TP - true positive;
        FP - false positive;
        FN - false negative;

    Args:
        gt: ground truth 4D keras tensor (B, H, W, C) or (B, C, H, W)
        pr: prediction 4D keras tensor (B, H, W, C) or (B, C, H, W)
        class_weights: 1. or list of class weights, len(weights) = C
        class_indexes: Optional integer or list of integers, classes to consider, if ``None`` all classes are used.
        beta: f-score coefficient
        smooth: value to avoid division by zero
        per_image: if ``True``, metric is calculated as mean over images in batch (B),
            else over whole batch
        threshold: value to round predictions (use ``>`` comparison), if ``None`` prediction will not be round

    Returns:
        F-score in range [0, 1]

    """

    backend = kwargs['backend']

    gt, pr = gather_channels(gt, pr, indexes=class_indexes, **kwargs)
    pr = round_if_needed(pr, threshold, **kwargs)
    axes = get_reduce_axes(per_image, **kwargs)

    # calculate score
    tp = backend.sum(gt * pr, axis=axes)
    fp = backend.sum(pr, axis=axes) - tp
    fn = backend.sum(gt, axis=axes) - tp

    score = ((1 + beta ** 2) * tp + smooth) \
            / ((1 + beta ** 2) * tp + beta ** 2 * fn + fp + smooth)
    score = average(score, per_image, class_weights, **kwargs)

    return score


def precision(gt, pr, class_weights=1, class_indexes=None, smooth=SMOOTH, per_image=False, threshold=None, **kwargs):
    r"""Calculate precision between the ground truth (gt) and the prediction (pr).

    .. math:: F_\beta(tp, fp) = \frac{tp} {(tp + fp)}

    where:
         - tp - true positives;
         - fp - false positives;

    Args:
        gt: ground truth 4D keras tensor (B, H, W, C) or (B, C, H, W)
        pr: prediction 4D keras tensor (B, H, W, C) or (B, C, H, W)
        class_weights: 1. or ``np.array`` of class weights (``len(weights) = num_classes``)
        class_indexes: Optional integer or list of integers, classes to consider, if ``None`` all classes are used.
        smooth: Float value to avoid division by zero.
        per_image: If ``True``, metric is calculated as mean over images in batch (B),
            else over whole batch.
        threshold: Float value to round predictions (use ``>`` comparison), if ``None`` prediction will not be round.
        name: Optional string, if ``None`` default ``precision`` name is used.

    Returns:
        float: precision score
    """
    backend = kwargs['backend']

    gt, pr = gather_channels(gt, pr, indexes=class_indexes, **kwargs)
    pr = round_if_needed(pr, threshold, **kwargs)
    axes = get_reduce_axes(per_image, **kwargs)

    # score calculation
    tp = backend.sum(gt * pr, axis=axes)
    fp = backend.sum(pr, axis=axes) - tp
    
    score = (tp + smooth) / (tp + fp + smooth)
    score = average(score, per_image, class_weights, **kwargs)

    return score


def recall(gt, pr, class_weights=1, class_indexes=None, smooth=SMOOTH, per_image=False, threshold=None, **kwargs):
    r"""Calculate recall between the ground truth (gt) and the prediction (pr).

    .. math:: F_\beta(tp, fn) = \frac{tp} {(tp + fn)}

    where:
         - tp - true positives;
         - fp - false positives;

    Args:
        gt: ground truth 4D keras tensor (B, H, W, C) or (B, C, H, W)
        pr: prediction 4D keras tensor (B, H, W, C) or (B, C, H, W)
        class_weights: 1. or ``np.array`` of class weights (``len(weights) = num_classes``)
        class_indexes: Optional integer or list of integers, classes to consider, if ``None`` all classes are used.
        smooth: Float value to avoid division by zero.
        per_image: If ``True``, metric is calculated as mean over images in batch (B),
            else over whole batch.
        threshold: Float value to round predictions (use ``>`` comparison), if ``None`` prediction will not be round.
        name: Optional string, if ``None`` default ``precision`` name is used.

    Returns:
        float: recall score
    """
    backend = kwargs['backend']

    gt, pr = gather_channels(gt, pr, indexes=class_indexes, **kwargs)
    pr = round_if_needed(pr, threshold, **kwargs)
    axes = get_reduce_axes(per_image, **kwargs)

    tp = backend.sum(gt * pr, axis=axes)
    fn = backend.sum(gt, axis=axes) - tp

    score = (tp + smooth) / (tp + fn + smooth)
    score = average(score, per_image, class_weights, **kwargs)

    return score


# ----------------------------------------------------------------
#   Loss Functions
# ----------------------------------------------------------------

def categorical_crossentropy(gt, pr, class_weights=1., class_indexes=None, **kwargs):
    backend = kwargs['backend']

    gt, pr = gather_channels(gt, pr, indexes=class_indexes, **kwargs)

    # scale predictions so that the class probas of each sample sum to 1
    axis = 3 if backend.image_data_format() == 'channels_last' else 1
    pr /= backend.sum(pr, axis=axis, keepdims=True)

    # clip to prevent NaN's and Inf's
    pr = backend.clip(pr, backend.epsilon(), 1 - backend.epsilon())

    # calculate loss
    output = gt * backend.log(pr) * class_weights
    return - backend.mean(output)


def binary_crossentropy(gt, pr, **kwargs):
    backend = kwargs['backend']
    return backend.mean(backend.binary_crossentropy(gt, pr))


def categorical_focal_loss(gt, pr, gamma=2.0, alpha=0.25, class_indexes=None, **kwargs):
    r"""Implementation of Focal Loss from the paper in multiclass classification

    Formula:
        loss = - gt * alpha * ((1 - pr)^gamma) * log(pr)

    Args:
        gt: ground truth 4D keras tensor (B, H, W, C) or (B, C, H, W)
        pr: prediction 4D keras tensor (B, H, W, C) or (B, C, H, W)
        alpha: the same as weighting factor in balanced cross entropy, default 0.25
        gamma: focusing parameter for modulating factor (1-p), default 2.0
        class_indexes: Optional integer or list of integers, classes to consider, if ``None`` all classes are used.

    """

    backend = kwargs['backend']
    gt, pr = gather_channels(gt, pr, indexes=class_indexes, **kwargs)

    # clip to prevent NaN's and Inf's
    pr = backend.clip(pr, backend.epsilon(), 1.0 - backend.epsilon())

    # Calculate focal loss
    loss = - gt * (alpha * backend.pow((1 - pr), gamma) * backend.log(pr))

    return backend.mean(loss)


def binary_focal_loss(gt, pr, gamma=2.0, alpha=0.25, **kwargs):
    r"""Implementation of Focal Loss from the paper in binary classification

    Formula:
        loss = - gt * alpha * ((1 - pr)^gamma) * log(pr) \
               - (1 - gt) * alpha * (pr^gamma) * log(1 - pr)

    Args:
        gt: ground truth 4D keras tensor (B, H, W, C) or (B, C, H, W)
        pr: prediction 4D keras tensor (B, H, W, C) or (B, C, H, W)
        alpha: the same as weighting factor in balanced cross entropy, default 0.25
        gamma: focusing parameter for modulating factor (1-p), default 2.0

    """
    backend = kwargs['backend']

    # clip to prevent NaN's and Inf's
    pr = backend.clip(pr, backend.epsilon(), 1.0 - backend.epsilon())

    loss_1 = - gt * (alpha * backend.pow((1 - pr), gamma) * backend.log(pr))
    loss_0 = - (1 - gt) * ((1 - alpha) * backend.pow((pr), gamma) * backend.log(1 - pr))
    loss = backend.mean(loss_0 + loss_1)
    return loss



# -*- coding: utf-8 -*-
"""
Created on Wed Aug 12 15:39:49 2020

@author: Koerber
"""

import tensorflow as tf
from dl.metric.metrics import Metrics, dlMetric
from dl.metric.metric_functions import *

class SegmentationMetrics(Metrics):
    def __init__(self,parent):
        super(SegmentationMetrics,self).__init__(parent)
        self.supportedmetrics.append(dlMetric.iou)
        self.supportedmetrics.append(dlMetric.dice_coeff)
        self.supportedmetrics.append(dlMetric.pixel_accuracy)
        self.metric = dlMetric.iou
        
    def getMetric(self):
        if self.metric == dlMetric.iou:
            return self.mean_iou_function(binary = self.parent.NumClasses <= 2, weighted = self.parent.seg_useWeightedDistanceMap)
        if self.metric == dlMetric.dice_coeff:
            return self.dice_coefficient_function(binary = self.parent.NumClasses <= 2, weighted = self.parent.seg_useWeightedDistanceMap)
        if self.metric == dlMetric.pixel_accuracy:
            return self.pixel_accuracy_function(binary = self.parent.NumClasses <= 2, weighted = self.parent.seg_useWeightedDistanceMap)
        else:
            raise # not supported
            
    def mean_iou_function(self, binary, weighted):
        #return [] # <- not running atm
        if binary:
            if weighted:
                return [mean_iou_binary_weighted]
            else:
                return [mean_iou_binary]
        else:
            if weighted:
                return [mean_iou_weighted]
            else:
                return [mean_iou]
         
    def dice_coefficient_function(self, binary, weighted):
        if binary:
            if weighted:
                return [dice_coefficient_binary_weighted]
            else:
                return [dice_coefficient_binary]
        else:
            if weighted:
                return [dice_coefficient_weighted]
            else:
                return [dice_coefficient]
        
    def pixel_accuracy_function(self, binary, weighted):
        if binary:
            if weighted:
                return [pixel_accuracy_binary_weighted]
            else:
                return [pixel_accuracy_binary]
        else:
            if weighted:
                return [pixel_accuracy_weighted]
            else:
                return [pixel_accuracy]


# -*- coding: utf-8 -*-
"""
Created on Wed Jul 22 13:49:54 2020

@author: Koerber
"""

from abc import ABC, abstractmethod
from enum import Enum
import numpy as np

class dlMode(Enum):
    Undefined = -1
    Classification = 0
    Object_Counting = 1
    Segmentation = 2
    # potentially at some point
    #Object_Detection = 3
    #Segmentation_3D = 4
    #Instance_Segmentation = 5

class dlPreprocess(Enum):
    scale = 0
    scale_shift = 1
    imagenet_mean = 2
    

class LearningMode(ABC):
    def __init__(self, parent):
        self.type = None
        self.parent = parent
        self.metric = None
        self.loss = None
        self.pretrained = True
        self.backbone = None
        self.architecture = None
        self.preprocessingfnc = None

    def preprocessImage(self, image):
        # should be set in case of using a pretrained model
        if self.preprocessingfnc:
            if image.dtype == np.uint16:
                image = image.astype('float')/255
            # float in range [0, 255] required
            return self.preprocessingfnc(image)

        if self.parent.preprocess == dlPreprocess.imagenet_mean:
            image = image.astype('float')
            if image.dtype == np.uint16:
                image = image.astype('float')/255
            
            # bgr mean of imagenet dataset
            mean = [103.939, 116.779, 123.68]
            if image.shape[-1] == 3:
                image[...,0] -= mean[0]
                image[...,1] -= mean[1]
                image[...,2] -= mean[2]
            elif image.shape[-1] == 1:
                image -= 127.5
            return image

        if image.dtype == np.uint16:
            scale_f = 65536
        else:
            scale_f = 255

        if self.parent.preprocess == dlPreprocess.scale:
            return image.astype('float')/scale_f
        if self.parent.preprocess == dlPreprocess.scale_shift:
            return image.astype('float')/(scale_f/2) -1
        
    @abstractmethod
    def LoadLabel(self, filename, height, width, rmBackground = False):
        pass
    
    @abstractmethod
    def prepreprocessLabel(self, label):
        # before training
        pass
        
    @abstractmethod
    def preprocessLabel(self, label):
        # during training
        pass

    @abstractmethod
    def setBackbone(self, bbname):
        return None

    @abstractmethod
    def setArchitecture(self, aname):
        return None
        
    @abstractmethod
    def getBackbones(self):
        return None

    @abstractmethod
    def getArchitectures(self):
        return None

    @abstractmethod
    def getModel(self, nclasses, channels):
        pass
    
    @abstractmethod
    def loadmodel(self, path):
        return None
    
    @abstractmethod
    def extractShapesFromPrediction(self, prediction, optArg1, optArg2):
        pass
    
    @abstractmethod
    def saveShapes(self, contours, path):
        pass
    
    @abstractmethod
    def PredictImage(self, image):
        pass
    
    @abstractmethod
    def resizeLabel(self, label, shape):
        pass
    
    @abstractmethod
    def countLabelWeight(self, label):
        pass

    @abstractmethod
    def getImageSize4ModelInput(self, inputWidth, inputHeight):
        pass
    
    @abstractmethod
    def setModelInputSize(self):
        pass


# -*- coding: utf-8 -*-
"""
Created on Wed Feb 19 13:43:00 2020

@author: Koerber
"""


from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *
from ui.ui_utils import LabelledAdaptiveDoubleSpinBox, LabelledSpinBox, LabelledDoubleSpinBox, DCDButton, openFolder
from ui.style import styleForm
from dl.optimizer.optimizer import dlOptim
from dl.training.lrschedule import lrMode
from dl.metric.metrics import dlMetric
from dl.loss.losses import dlLoss
from dl.method.mode import dlPreprocess


class Window(object):
    def setupUi(self, Form):

        Form.setWindowTitle('Training Settings') 
        styleForm(Form)

        self.centralWidget = QWidget(Form)

        self.vlayout = QVBoxLayout(self.centralWidget)
        self.vlayout.setContentsMargins(3, 3, 3, 3)
        self.vlayout.setSpacing(6)
        self.centralWidget.setLayout(self.vlayout)

        layout = QHBoxLayout(self.centralWidget)
        self.CBMemory = QCheckBox("Load Full Dataset",self.centralWidget)
        self.CBMemory.setToolTip('Check to load full dataset into memory, uncheck to reload data on each iteration')
        self.CBMemory.setObjectName('nn_TrainInMemory')
        layout.addWidget(self.CBMemory)

        self.SBPredictEvery = LabelledSpinBox ('Predict an image every x epochs', self.centralWidget)
        self.SBPredictEvery.setToolTip('Predict the current selected test image after every x epochs')
        self.SBPredictEvery.SpinBox.setRange(0,100)
        self.SBPredictEvery.SpinBox.setValue(0)
        self.SBPredictEvery.setObjectName('nn_PredictEveryX')
        self.CBPreprocess = QComboBox(self.centralWidget)
        self.CBPreprocess.setObjectName('nn_Preprocessing')
        for pre in dlPreprocess:
            self.CBPreprocess.addItem(pre.name)

        layout.addWidget(self.CBPreprocess)
        self.LPreprocess = QLabel('Preprocessing')
        layout.addWidget(self.LPreprocess)

        self.vlayout.addLayout(layout)
        self.vlayout.addWidget(self.SBPredictEvery)
        
        self.SBBorder = LabelledSpinBox('Ignore border of training images',self.centralWidget)
        self.SBBorder.setObjectName('nn_IgnoreBorder')
        self.SBBorder.setToolTip('Set to remove a border in pixel around each training image')
        self.SBBorder.SpinBox.setRange(0,100)
        self.vlayout.addWidget(self.SBBorder) 
        
        
        self.vlayout.addWidget(self.LossGroup())
        self.vlayout.addWidget(self.TrainValGroup())
        self.vlayout.addWidget(self.LRScheduleGroup())

        self.centralWidget.setFixedSize(self.vlayout.sizeHint())
        Form.setFixedSize(self.vlayout.sizeHint())

    def LossGroup(self):
        groupBox = QGroupBox("Loss, Metric, Optimizer")
        
        layout = QVBoxLayout(self.centralWidget)
        hlayout = QHBoxLayout(self.centralWidget)

        hlayout_2 = QHBoxLayout(self.centralWidget)
        
        self.LWeights = QLabel("Class Weighting",self.centralWidget)
        self.RBAutoClassWeight = QRadioButton("Auto weighting",self.centralWidget)
        self.RBAutoClassWeight.setToolTip('Check to automatically calculate the weight of each class')
        self.RBAutoClassWeight.setObjectName('nn_AutoClassWeight')
        
        self.RBManuClassWeight = QRadioButton("Manual weighting",self.centralWidget)
        self.RBManuClassWeight.setToolTip('Check to manual set weight of each class')
        self.RBManuClassWeight.setObjectName('nn_ManualClassWeight')
        
        self.RBNoClassWeight = QRadioButton("Disable weighting",self.centralWidget)
        self.RBNoClassWeight.setToolTip('Check to disable class weights')
        self.RBNoClassWeight.setObjectName('nn_NoClassWeight')
        
        hlayout_2.addWidget(self.RBNoClassWeight)
        hlayout_2.addWidget(self.RBAutoClassWeight)
        hlayout_2.addWidget(self.RBManuClassWeight)
        
        hlayout_3 = QHBoxLayout(self.centralWidget)
        
        self.CBClassWeights = QComboBox(self.centralWidget)
        self.SBWeight = LabelledAdaptiveDoubleSpinBox ('Class weight', self.centralWidget)
        self.SBWeight.SpinBox.setRange(0.0001,1)
        self.SBWeight.setToolTip('Set class weight')
        self.SBWeight.SpinBox.setDecimals(4)
        self.SBWeight.SpinBox.setSingleStep(0.1)
        hlayout_3.addWidget(self.CBClassWeights)
        hlayout_3.addWidget(self.SBWeight)

        layout.addWidget(self.LWeights)
        layout.addLayout(hlayout_2)
        layout.addLayout(hlayout_3)
        
        vlayout_1 = QVBoxLayout(self.centralWidget)
        self.LLoss = QLabel("Loss",self.centralWidget)
        self.CBLoss = QComboBox(self.centralWidget)
        self.CBLoss.setObjectName('nn_Loss')
        self.CBLoss.setToolTip('Set loss function that is optimized during training')
        vlayout_1.addWidget(self.LLoss)
        vlayout_1.addWidget(self.CBLoss)
        
        vlayout_2 = QVBoxLayout(self.centralWidget)
        self.LMetric = QLabel("Metric",self.centralWidget)
        self.CBMetrics = QComboBox(self.centralWidget)
        self.CBMetrics.setObjectName('nn_Metric')
        self.CBMetrics.setToolTip('Set metric to measure network performance')
        vlayout_2.addWidget(self.LMetric)
        vlayout_2.addWidget(self.CBMetrics)
        vlayout_3 = QVBoxLayout(self.centralWidget)
        self.LOptimizer = QLabel("Optimizer",self.centralWidget)
        self.CBOptimizer = QComboBox(self.centralWidget)
        self.CBOptimizer.setObjectName('nn_Optimizer')
        vlayout_3.addWidget(self.LOptimizer)
        vlayout_3.addWidget(self.CBOptimizer)
        
        
        for opt in dlOptim:
            self.CBOptimizer.addItem(opt.name)
        self.CBOptimizer.setToolTip('Set optimizer to minimize loss function')
        hlayout.addLayout(vlayout_1)
        hlayout.addLayout(vlayout_2)
        hlayout.addLayout(vlayout_3)

        layout.addLayout(hlayout)
        
        groupBox.setLayout(layout)

        return groupBox

    def TrainValGroup(self):
        groupBox = QGroupBox("Train/Validation Split")
        
        vlayout = QVBoxLayout(self.centralWidget)
        h1layout = QHBoxLayout(self.centralWidget)
        h2layout = QHBoxLayout(self.centralWidget)
        
        self.RBUseDir = QRadioButton(self.centralWidget)
        self.RBUseDir.setObjectName('ValidationDir')
        self.RBUseDir.setToolTip('Toggle to use a validation directory')
        
        self.BLoadValidationImages = DCDButton(self.centralWidget)
        self.BLoadValidationImages.setToolTip('Select folder with validation images')
        self.BLoadValidationImages.setIcon(QIcon('icons/load.png'))
        
        self.LVal = QLabel(self.centralWidget)
        self.LVal.setText('Validation Folder')
        
        h1layout.addWidget(self.RBUseDir ,stretch=1)
        h1layout.addWidget(self.BLoadValidationImages, stretch=1)
        h1layout.addWidget(self.LVal, stretch=99)
        
        
        self.RBUseSplit = QRadioButton(self.centralWidget)
        self.RBUseSplit.setObjectName('ValidationSplit')
        self.RBUseSplit.setToolTip('Toggle to use a validation split')
        self.STrainVal = QSlider(Qt.Horizontal, self.centralWidget)
        self.STrainVal.setObjectName('nn_TrainValSplit')
        self.STrainVal.setMinimum(20)
        self.STrainVal.setMaximum(100)
        self.STrainVal.setValue(100)
        self.STrainVal.setToolTip('Split training data into a train and validation set.')
        self.LTrainVal = QLabel(self.centralWidget)
        self.LTrainVal.setText('100 % Train 0 % Val')
        

        
        h2layout.addWidget(self.RBUseSplit)
        h2layout.addWidget(self.STrainVal)
        h2layout.addWidget(self.LTrainVal)

        vlayout.addLayout(h1layout)
        vlayout.addLayout(h2layout)
        groupBox.setLayout(vlayout)

        return groupBox

    def LRScheduleGroup(self):
        groupBox = QGroupBox("Learning Rate Schedule")
        
        layout = QVBoxLayout(self.centralWidget)
        hlayout = QHBoxLayout(self.centralWidget)

        self.RBConstant = QRadioButton(self.centralWidget)
        self.RBConstant.setObjectName('nn_ConstantLR')
        self.RBConstant.setText("Constant")
        self.RBConstant.setToolTip('Toggle to set constant learning rate')


        self.RBRlroP = QRadioButton(self.centralWidget)
        self.RBRlroP.setObjectName('nn_ReduceLR')
        self.RBRlroP.setText("on Plateau")
        self.RBRlroP.setToolTip('Toggle to reduce learning rate on plateau')

        self.RBLRschedule = QRadioButton(self.centralWidget)
        self.RBLRschedule.setObjectName('nn_ScheduleLR')
        self.RBLRschedule.setText("Schedule")
        self.RBLRschedule.setToolTip('Toggle to set learning rate schedule')

        hlayout.addWidget(self.RBConstant)
        hlayout.addWidget(self.RBRlroP)
        hlayout.addWidget(self.RBLRschedule)
        layout.addLayout(hlayout)

        self.SBlr = LabelledAdaptiveDoubleSpinBox ('Learning Rate', self.centralWidget)
        self.SBlr.setObjectName('nn_Learning_Rate')
        self.SBlr.SpinBox.setRange(0.000001,1)
        self.SBlr.setToolTip('Set learning rate')
        self.SBlr.SpinBox.setDecimals(6)
        self.SBlr.SpinBox.setSingleStep(0.1)
        layout.addWidget(self.SBlr)
        
        self.SBReduction = LabelledAdaptiveDoubleSpinBox ('Reduction Factor', self.centralWidget)
        self.SBReduction.setObjectName('nn_LRReduction')
        self.SBReduction.SpinBox.setRange(0.001,1)
        self.SBReduction.SpinBox.setDecimals(3)
        self.SBReduction.SpinBox.setSingleStep(0.1)
        layout.addWidget(self.SBReduction)

        self.SBEveryx = LabelledSpinBox ('Reduce every x epochs', self.centralWidget)
        self.SBEveryx.setObjectName('nn_LREpochs')
        self.SBEveryx.SpinBox.setRange(1,1000)
        self.SBEveryx.SpinBox.setValue(25)
        layout.addWidget(self.SBEveryx)
        layout.addStretch()

        groupBox.setLayout(layout)

        return groupBox


class TrainSettingsWindow(QMainWindow, Window):
    def __init__(self, parent ):
        super(TrainSettingsWindow, self).__init__(parent)
        self.parent = parent
        self.setupUi(self)
        
        self.CBMemory.setChecked(self.parent.parent.dl.TrainInMemory)
        
        self.CBOptimizer.setCurrentIndex(self.CBOptimizer.findText(self.parent.parent.dl.optimizer.Optimizer.name))
        self.SBlr.SpinBox.setValue(self.parent.parent.dl.learning_rate)
        self.SBlr.SpinBox.setMinimum(self.parent.parent.dl.lrschedule.minlr)
        self.SBPredictEvery.SpinBox.setValue(self.parent.parent.predictionRate)
        
        self.CBMemory.stateChanged.connect(self.InMemoryChanged)
        
        self.RBNoClassWeight.setChecked(not self.parent.parent.dl.data.autocalcClassWeights)
        self.RBNoClassWeight.toggled.connect(self.changeClassWeightSettings)
        self.RBAutoClassWeight.toggled.connect(self.changeClassWeightSettings)
        self.RBManuClassWeight.toggled.connect(self.changeClassWeightSettings)
        self.CBClassWeights.currentIndexChanged.connect(self.changeSelectedClass)
        self.SBWeight.SpinBox.valueChanged.connect(self.changeClassWeights)

        self.CBOptimizer.currentIndexChanged.connect(self.changeOptimizer)
        self.CBMetrics.currentIndexChanged.connect(self.changeMetric)
        self.CBLoss.currentIndexChanged.connect(self.changeLoss)
        self.CBPreprocess.setCurrentIndex(self.CBOptimizer.findText(self.parent.parent.dl.preprocess.name))
        self.CBPreprocess.currentIndexChanged.connect(self.changePreprocessing)
        
        self.SBBorder.SpinBox.setValue(self.parent.parent.dl.data.IgnoreBorder)
        self.SBBorder.SpinBox.valueChanged.connect(self.changeBorder)

        self.RBConstant.toggled.connect(self.LROption)
        self.RBRlroP.toggled.connect(self.LROption)
        self.RBLRschedule.toggled.connect(self.LROption)
        
        self.RBUseDir.toggled.connect(self.valsetChanged)
        self.RBUseSplit.toggled.connect(self.valsetChanged)
        self.BLoadValidationImages.clicked.connect(self.selectValDir)
        self.RBUseDir.setChecked(self.parent.parent.dl.data.UseValidationDir)
        
        
        self.SBlr.SpinBox.valueChanged.connect(self.changeLR)
        self.SBReduction.SpinBox.valueChanged.connect(self.changeReductionFactor)
        self.SBEveryx.SpinBox.valueChanged.connect(self.changeIntervall)
        self.SBPredictEvery.SpinBox.valueChanged.connect(self.predictionRateChanged)
        
        self.STrainVal.valueChanged.connect(self.changeTrainTestSplit)
        self.STrainVal.setValue(self.parent.parent.dl.data.TrainTestSplit*100)
        
        self.RBRlroP.setChecked(True)
        self.LROption()
        self.updateLossesAndMetrics()
        self.changeClassWeightSettings()

    def InMemoryChanged(self):
        self.parent.parent.dl.TrainInMemory = self.CBMemory.isChecked()
        
    def predictionRateChanged(self):
        self.parent.parent.predictionRate = self.SBPredictEvery.SpinBox.value()
        
    def changeOptimizer(self, i):
        self.parent.parent.dl.optimizer.setOptimizer(dlOptim[self.CBOptimizer.currentText()])
        self.parent.SBLearningRate.SpinBox.setValue(self.parent.parent.dl.learning_rate)
        
    def LROption(self):
        if self.RBConstant.isChecked():
            self.SBReduction.hide()
            self.SBEveryx.hide()
            self.parent.parent.dl.lrschedule.setMode(lrMode.Constant)
        if self.RBRlroP.isChecked():
            self.SBReduction.show()
            self.SBEveryx.show()
            self.SBEveryx.Label.setText('Plateau for x epochs')   
            self.parent.parent.dl.lrschedule.setMode(lrMode.ReduceOnPlateau)
        if self.RBLRschedule.isChecked():
            self.SBReduction.show()
            self.SBEveryx.show()
            self.SBEveryx.Label.setText('Reduce every x epochs')
            self.parent.parent.dl.lrschedule.setMode(lrMode.Schedule)
        self.SBEveryx.SpinBox.setValue(self.parent.parent.dl.lrschedule.ReduceEveryXEpoch)
        self.SBReduction.SpinBox.setValue(self.parent.parent.dl.lrschedule.ReductionFactor)
            
    def changeReductionFactor(self):
        self.parent.parent.dl.lrschedule.ReductionFactor = self.SBReduction.SpinBox.value() 
        
    def changeTrainTestSplit(self):
        self.parent.parent.dl.data.TrainTestSplit = self.STrainVal.value() / 100.
        text = '%d %% Train  %d %% Val' % (self.parent.parent.dl.data.TrainTestSplit*100, 100-self.parent.parent.dl.data.TrainTestSplit*100)
        self.LTrainVal.setText(text)
        
    def changeIntervall(self):
        self.parent.parent.dl.lrschedule.ReduceEveryXEpoch = self.SBEveryx.SpinBox.value()
 
    def changeLR(self):
        self.parent.SBLearningRate.SpinBox.setValue(self.SBlr.SpinBox.value())
            
    def silentlyUpdateLR(self, value):
        self.SBlr.SpinBox.blockSignals(True)
        self.SBlr.SpinBox.setValue(value)
        self.SBlr.SpinBox.blockSignals(False)

    def changeLoss(self):
        if self.CBLoss.currentText() != '': 
            self.parent.parent.dl.Mode.loss.loss = dlLoss[self.CBLoss.currentText()]

    def changePreprocessing(self):
        if self.CBPreprocess.currentText() != '': 
            self.parent.parent.dl.preprocess = dlPreprocess[self.CBPreprocess.currentText()]

    def changeMetric(self):
        if self.CBMetrics.currentText() != '':
            self.parent.parent.dl.Mode.metric.metric = dlMetric[self.CBMetrics.currentText()]

    def updateLossesAndMetrics(self):
        self.CBLoss.clear()
        self.CBMetrics.clear()
        default_loss = self.parent.parent.dl.Mode.loss.loss.name
        default_metric = self.parent.parent.dl.Mode.metric.metric.name
        for loss in self.parent.parent.dl.Mode.loss.supportedlosses:
            self.CBLoss.addItem(loss.name)
        for metric in self.parent.parent.dl.Mode.metric.supportedmetrics:
            self.CBMetrics.addItem(metric.name)
        self.CBLoss.setCurrentIndex(self.CBLoss.findText(default_loss))
        self.CBMetrics.setCurrentIndex(self.CBMetrics.findText(default_metric))

    def changeClassWeightSettings(self):
        self.parent.parent.dl.data.autocalcClassWeights = self.RBAutoClassWeight.isChecked()
        self.CBClassWeights.clear()
        
        if self.parent.parent.NumOfClasses() > 2:
            for c in range(self.parent.parent.NumOfClasses()):
                self.CBClassWeights.addItem(self.parent.parent.classList.getClassName(c))
        else:
            self.CBClassWeights.addItem(self.parent.parent.classList.getClassName(0))
            self.CBClassWeights.addItem(self.parent.parent.classList.getClassName(1))
        self.CBClassWeights.setCurrentIndex(1) 
        if self.RBManuClassWeight.isChecked():
            self.CBClassWeights.setEnabled(True)
            self.SBWeight.setEnabled(True)
            self.changeClassWeights()
        else:
            self.CBClassWeights.setEnabled(False)
            self.SBWeight.setEnabled(False)
            
        if self.RBNoClassWeight.isChecked():
            self.parent.parent.dl.data.resetClassWeights()
            
    def changeSelectedClass(self):
        if self.CBClassWeights.currentIndex() >= 0:
            self.SBWeight.SpinBox.setValue(self.parent.parent.classList.getClassWeight(self.CBClassWeights.currentIndex()))
            
    def changeClassWeights(self):
        self.parent.parent.classList.setClassWeight(self.CBClassWeights.currentIndex(),self.SBWeight.SpinBox.value())
        if self.parent.parent.NumOfClasses() == 2:
            idx = 1 - self.CBClassWeights.currentIndex()
            self.parent.parent.classList.setClassWeight(idx,1-self.SBWeight.SpinBox.value())
        self.parent.parent.setClassWeights()

            
    def valsetChanged(self):
        self.parent.parent.dl.data.UseValidationDir = self.RBUseDir.isChecked()
        self.STrainVal.setEnabled(self.RBUseSplit.isChecked())
        self.BLoadValidationImages.setEnabled(self.RBUseDir.isChecked())
        
        

    def selectValDir(self):
        folder = openFolder("Select Validation Image Folder") 
        if folder:
            self.parent.parent.files.validationpath = folder
            self.LVal.setText(folder)
        
    def changeBorder(self):
        self.parent.parent.dl.data.IgnoreBorder = self.SBBorder.SpinBox.value()



from setuptools import setup, find_packages
import os



here = os.path.abspath(os.path.dirname(__file__))
ver = {} 
with open(os.path.join(here, 'mia', '__version__.py')) as f:
    exec(f.read(), ver)

setup(
    name="mianalyzer", 
    version=ver['__version__'],
    author="Nils Koerber",
    description="MIA deep learning based Microscopic Image Analyzer",
	long_description="MIA is a software for deep learning based image analysis. It covers image labeling, neural network training and inference. It can be used for image classification, object detection, semantic segmentation and tracking, see https://github.com/MIAnalyzer/MIA/ for details.",
    url="https://github.com/MIAnalyzer/MIA/",

	install_requires=[
	# requirements removed here as MIA installation via conda is recommended
    # when installing via pip use requirements.txt
	],

    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
	packages=find_packages(include=[
        "mia",
        "mia.*"
    ]),
	entry_points={
        'console_scripts': [
            'mianalyzer = mia.startup:start',
        ],
        # downloads do not work without console -> fix
		#'gui_scripts': [
        #    'mia = mia.startup:start',
        #],
	},
	data_files=[('icons', [
	'mia/icons/addobject.png', 
	'mia/icons/assign.png',
	'mia/icons/assignclass.png', 
	'mia/icons/augmentation.png',
	'mia/icons/clear.png', 
	'mia/icons/delete.png',
	'mia/icons/dextr.png', 
	'mia/icons/drag.png',
	'mia/icons/draw.png', 
	'mia/icons/expand.png',
	'mia/icons/exportallmasks.png', 
	'mia/icons/exportmask.png',
	'mia/icons/load.png', 
	'mia/icons/loadmodel.png',
	'mia/icons/magicwand.png',
	'mia/icons/measure.png', 
	'mia/icons/next.png',
	'mia/icons/objectcolor.png',
	'mia/icons/objectnumber.png', 
	'mia/icons/poly.png',
	'mia/icons/postprocessing.png', 
	'mia/icons/predict.png',
	'mia/icons/predictall.png', 
	'mia/icons/previous.png',
	'mia/icons/reset.png',
	'mia/icons/results.png', 
	'mia/icons/saveall.png',
	'mia/icons/savemodel.png', 
	'mia/icons/setclass.png',
	'mia/icons/settings.png', 
	'mia/icons/shift.png',
	'mia/icons/tracking.png', 
	'mia/icons/train.png',
	'mia/icons/logo.png', 
	'mia/icons/loading.png', 
	'mia/dl/machine_learning/deploy.prototxt',
	]
	
	),
     ],

	include_package_data=True,
	
    python_requires=">=3.6",
)


import os
import collections

from dl.models.classification_models import get_submodules_from_kwargs
from ._common_blocks import GroupConv2D
from ..weights import load_model_weights

backend = None
layers = None
models = None
keras_utils = None

ModelParams = collections.namedtuple(
    'ModelParams',
    ['model_name', 'repetitions']
)


# -------------------------------------------------------------------------
#   Helpers functions
# -------------------------------------------------------------------------

def handle_block_names(stage, block):
    name_base = 'stage{}_unit{}_'.format(stage + 1, block + 1)
    conv_name = name_base + 'conv'
    bn_name = name_base + 'bn'
    relu_name = name_base + 'relu'
    sc_name = name_base + 'sc'
    return conv_name, bn_name, relu_name, sc_name


def get_conv_params(**params):
    default_conv_params = {
        'kernel_initializer': 'glorot_uniform',
        'use_bias': False,
        'padding': 'valid',
    }
    default_conv_params.update(params)
    return default_conv_params


def get_bn_params(**params):
    axis = 3 if backend.image_data_format() == 'channels_last' else 1
    default_bn_params = {
        'axis': axis,
        'momentum': 0.99,
        'epsilon': 2e-5,
        'center': True,
        'scale': True,
    }
    default_bn_params.update(params)
    return default_bn_params


# -------------------------------------------------------------------------
#   Residual blocks
# -------------------------------------------------------------------------


def conv_block(filters, stage, block, strides=(2, 2), **kwargs):
    """The conv block is the block that has conv layer at shortcut.
    # Arguments
        filters: integer, used for first and second conv layers, third conv layer double this value
        strides: tuple of integers, strides for conv (3x3) layer in block
        stage: integer, current stage label, used for generating layer names
        block: integer, current block label, used for generating layer names
    # Returns
        Output layer for the block.
    """

    def layer(input_tensor):
        # extracting params and names for layers
        conv_params = get_conv_params()
        group_conv_params = dict(list(conv_params.items()) + list(kwargs.items()))
        bn_params = get_bn_params()
        conv_name, bn_name, relu_name, sc_name = handle_block_names(stage, block)

        x = layers.Conv2D(filters, (1, 1), name=conv_name + '1', **conv_params)(input_tensor)
        x = layers.BatchNormalization(name=bn_name + '1', **bn_params)(x)
        x = layers.Activation('relu', name=relu_name + '1')(x)

        x = layers.ZeroPadding2D(padding=(1, 1))(x)
        x = GroupConv2D(filters, (3, 3), strides=strides, **group_conv_params)(x)
        x = layers.BatchNormalization(name=bn_name + '2', **bn_params)(x)
        x = layers.Activation('relu', name=relu_name + '2')(x)

        x = layers.Conv2D(filters * 2, (1, 1), name=conv_name + '3', **conv_params)(x)
        x = layers.BatchNormalization(name=bn_name + '3', **bn_params)(x)

        shortcut = layers.Conv2D(filters * 2, (1, 1), name=sc_name, strides=strides, **conv_params)(input_tensor)
        shortcut = layers.BatchNormalization(name=sc_name + '_bn', **bn_params)(shortcut)
        x = layers.Add()([x, shortcut])

        x = layers.Activation('relu', name=relu_name)(x)
        return x

    return layer


def identity_block(filters, stage, block, **kwargs):
    """The identity block is the block that has no conv layer at shortcut.
    # Arguments
        filters: integer, used for first and second conv layers, third conv layer double this value
        stage: integer, current stage label, used for generating layer names
        block: integer, current block label, used for generating layer names
    # Returns
        Output layer for the block.
    """

    def layer(input_tensor):
        conv_params = get_conv_params()
        group_conv_params = dict(list(conv_params.items()) + list(kwargs.items()))
        bn_params = get_bn_params()
        conv_name, bn_name, relu_name, sc_name = handle_block_names(stage, block)

        x = layers.Conv2D(filters, (1, 1), name=conv_name + '1', **conv_params)(input_tensor)
        x = layers.BatchNormalization(name=bn_name + '1', **bn_params)(x)
        x = layers.Activation('relu', name=relu_name + '1')(x)

        x = layers.ZeroPadding2D(padding=(1, 1))(x)
        x = GroupConv2D(filters, (3, 3), **group_conv_params)(x)
        x = layers.BatchNormalization(name=bn_name + '2', **bn_params)(x)
        x = layers.Activation('relu', name=relu_name + '2')(x)

        x = layers.Conv2D(filters * 2, (1, 1), name=conv_name + '3', **conv_params)(x)
        x = layers.BatchNormalization(name=bn_name + '3', **bn_params)(x)

        x = layers.Add()([x, input_tensor])

        x = layers.Activation('relu', name=relu_name)(x)
        return x

    return layer


def ResNeXt(
        model_params,
        include_top=True,
        input_tensor=None,
        input_shape=None,
        classes=1000,
        weights='imagenet',
        **kwargs):
    """Instantiates the ResNet, SEResNet architecture.
    Optionally loads weights pre-trained on ImageNet.
    Note that the data format convention used by the model is
    the one specified in your Keras config at `~/.keras/keras.json`.

    Args:
        include_top: whether to include the fully-connected
            layer at the top of the network.
        weights: one of `None` (random initialization),
              'imagenet' (pre-training on ImageNet),
              or the path to the weights file to be loaded.
        input_tensor: optional Keras tensor
            (i.e. output of `layers.Input()`)
            to use as image input for the model.
        input_shape: optional shape tuple, only to be specified
            if `include_top` is False (otherwise the input shape
            has to be `(224, 224, 3)` (with `channels_last` data format)
            or `(3, 224, 224)` (with `channels_first` data format).
            It should have exactly 3 inputs channels.
        classes: optional number of classes to classify images
            into, only to be specified if `include_top` is True, and
            if no `weights` argument is specified.

    Returns:
        A Keras model instance.

    Raises:
        ValueError: in case of invalid argument for `weights`,
            or invalid input shape.
    """

    global backend, layers, models, keras_utils
    backend, layers, models, keras_utils = get_submodules_from_kwargs(kwargs)

    if input_tensor is None:
        img_input = layers.Input(shape=input_shape, name='data')
    else:
        if not backend.is_keras_tensor(input_tensor):
            img_input = layers.Input(tensor=input_tensor, shape=input_shape)
        else:
            img_input = input_tensor

    # get parameters for model layers
    no_scale_bn_params = get_bn_params(scale=False)
    bn_params = get_bn_params()
    conv_params = get_conv_params()

    # resnext bottom
    x = layers.BatchNormalization(name='bn_data', **no_scale_bn_params)(img_input)
    x = layers.ZeroPadding2D(padding=(3, 3))(x)
    x = layers.Conv2D(64, (7, 7), strides=(2, 2), name='conv0', **conv_params)(x)
    x = layers.BatchNormalization(name='bn0', **bn_params)(x)
    x = layers.Activation('relu', name='relu0')(x)
    x = layers.ZeroPadding2D(padding=(1, 1))(x)
    x = layers.MaxPooling2D((3, 3), strides=(2, 2), padding='valid', name='pooling0')(x)

    # resnext body
    init_filters = 128
    for stage, rep in enumerate(model_params.repetitions):
        for block in range(rep):

            filters = init_filters * (2 ** stage)

            # first block of first stage without strides because we have maxpooling before
            if stage == 0 and block == 0:
                x = conv_block(filters, stage, block, strides=(1, 1), **kwargs)(x)

            elif block == 0:
                x = conv_block(filters, stage, block, strides=(2, 2), **kwargs)(x)

            else:
                x = identity_block(filters, stage, block, **kwargs)(x)

    # resnext top
    if include_top:
        x = layers.GlobalAveragePooling2D(name='pool1')(x)
        x = layers.Dense(classes, name='fc1')(x)
        x = layers.Activation('softmax', name='softmax')(x)

    # Ensure that the model takes into account any potential predecessors of `input_tensor`.
    if input_tensor is not None:
        inputs = keras_utils.get_source_inputs(input_tensor)
    else:
        inputs = img_input

    # Create model
    model = models.Model(inputs, x)

    if weights:
        if type(weights) == str and os.path.exists(weights):
            model.load_weights(weights)
        else:
            load_model_weights(model, model_params.model_name,
                               weights, classes, include_top, **kwargs)

    return model


# -------------------------------------------------------------------------
#   Residual Models
# -------------------------------------------------------------------------

MODELS_PARAMS = {
    'resnext50': ModelParams('resnext50', (3, 4, 6, 3)),
    'resnext101': ModelParams('resnext101', (3, 4, 23, 3)),
}


def ResNeXt50(input_shape=None, input_tensor=None, weights=None, classes=1000, include_top=True, **kwargs):
    return ResNeXt(
        MODELS_PARAMS['resnext50'],
        input_shape=input_shape,
        input_tensor=input_tensor,
        include_top=include_top,
        classes=classes,
        weights=weights,
        **kwargs
    )


def ResNeXt101(input_shape=None, input_tensor=None, weights=None, classes=1000, include_top=True, **kwargs):
    return ResNeXt(
        MODELS_PARAMS['resnext101'],
        input_shape=input_shape,
        input_tensor=input_tensor,
        include_top=include_top,
        classes=classes,
        weights=weights,
        **kwargs
    )


def preprocess_input(x, **kwargs):
    return x


setattr(ResNeXt50, '__doc__', ResNeXt.__doc__)
setattr(ResNeXt101, '__doc__', ResNeXt.__doc__)



import threading

class WorkerThread(threading.Thread):
    def __init__(self, work, args=None, callback_start=None,callback_end=None):
        threading.Thread.__init__(self)
        self.work = work
        self.callback_end = callback_end
        self.callback_start = callback_start
        self.arguments = args
        
    def run(self):
        if self.callback_start:
            self.callback_start()
        if self.arguments:
            self.work(self.arguments)
        else:
            self.work()
        if self.callback_end:
            self.callback_end()






from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *

def setStyle(app):
    # # Force the style to be the same on all OSs:
    app.setStyle("Fusion")
    


    # # Now use a palette to switch to dark colors:
    palette = QPalette()
    palette.setColor(QPalette.Window, getBackgroundColor())
    palette.setColor(QPalette.WindowText, Qt.white)
    palette.setColor(QPalette.Disabled, QPalette.WindowText, Qt.gray)
    
    
    palette.setColor(QPalette.Base, QColor(25, 25, 25))
    palette.setColor(QPalette.AlternateBase, getBrightColor())
    # have no effect, see below
    palette.setColor(QPalette.ToolTipBase, getBackgroundColor())
    palette.setColor(QPalette.ToolTipText, getHighlightColor())

    palette.setColor(QPalette.Text, Qt.white)
    palette.setColor(QPalette.Button, getBackgroundColor())
    
    palette.setColor(QPalette.ButtonText, Qt.white)
    palette.setColor(QPalette.BrightText, Qt.red)


    palette.setColor(QPalette.Highlight,QColor(51,153,255))
    palette.setColor(QPalette.HighlightedText, getHighlightColor())
    
    # every time a tooltip is set, the bg-color resets
    # so background-color here and border do not have an effect
    app.setStyleSheet("""QToolTip { background-color: QColor(53, 53, 53); color: lightblue; border: lightblue solid 1px }""")  
    app.setPalette(palette)
    
    
def getBrightColor(asString = False):
    color = QColor(80, 80, 80)
    if asString:
        return "rgb(%d, %d, %d)" % (color.red(), color.green(), color.blue())
    return color

def getHighlightColor(asString = False):
    color = QColor(173, 216, 230)
    if asString:
        return "rgb(%d, %d, %d)" % (color.red(), color.green(), color.blue())
    return color

def getBackgroundColor(asString = False):
    color = QColor(53, 53, 53)
    if asString:
        return "rgb(%d, %d, %d)" % (color.red(), color.green(), color.blue())
    return color


def styleForm(form):
    # add something like:
    #form.setStyleSheet("QRadioButton { color: red }")
#    palette = QPalette()
#    palette.setColor(QPalette.ToolTipBase, Qt.white)
#    palette.setColor(QPalette.ToolTipText, Qt.white)
#    form.setPalette(palette)
    pass 



# -*- coding: utf-8 -*-
"""
Created on Wed Aug 12 14:30:51 2020

@author: Koerber
"""

import tensorflow as tf
from dl.loss.losses import Losses, dlLoss
from dl.loss.loss_functions import *

class SegmentationLosses(Losses):
    def __init__(self,parent):
        super(SegmentationLosses,self).__init__(parent)
        self.supportedlosses.append(dlLoss.focal_loss)
        self.supportedlosses.append(dlLoss.cross_entropy)
        self.supportedlosses.append(dlLoss.dice_loss)
        self.supportedlosses.append(dlLoss.kl_divergence)
        self.loss = dlLoss.focal_loss
        
    def getLoss(self):
        if self.loss == dlLoss.focal_loss:
            return self.focal_Loss_function(binary = self.parent.NumClasses <= 2, usedist = self.parent.seg_useWeightedDistanceMap, class_weights=self.parent.data.class_weights)
        if self.loss == dlLoss.cross_entropy:
            return self.categorical_cross_entropy_loss_function(binary = self.parent.NumClasses <= 2, usedist = self.parent.seg_useWeightedDistanceMap, class_weights=self.parent.data.class_weights)
        if self.loss == dlLoss.dice_loss:
            return self.dice_coeffient_loss_function(binary = self.parent.NumClasses <= 2, usedist = self.parent.seg_useWeightedDistanceMap, class_weights=self.parent.data.class_weights)
        if self.loss == dlLoss.kl_divergence:
            return self.kullback_Leibler_divergence_loss_function(usedist = self.parent.seg_useWeightedDistanceMap)
        else:
            raise # not supported
                    
        
    def categorical_cross_entropy_loss_function(self, binary, usedist,class_weights):
        # side note: in retina-net paper alpha = 0.75 gives best result for cce
        if class_weights is None:
            class_weights = 0.5

        if binary:
            return focal_loss_binary(usedistmap=usedist, class_weights=class_weights, gamma=0)
        else:
            return focal_loss(usedistmap=usedist, class_weights=class_weights, gamma=0)

    def focal_Loss_function(self, binary, usedist, class_weights):
        if binary:
            return focal_loss_binary(usedistmap=usedist, class_weights=class_weights, gamma=2)
        else:
            return focal_loss(usedistmap=usedist, class_weights=class_weights, gamma=2)
    
    def kullback_Leibler_divergence_loss_function(self, usedist):
        return kullback_Leibler_divergence_loss_weighted(usedist)
    
    def dice_coeffient_loss_function(self, binary, usedist, class_weights):
        if binary:
            return jaccard_distance_loss_binary(usedistmap=usedist, class_weights=class_weights)
        else:
            return jaccard_distance_loss(usedistmap=usedist, class_weights=class_weights)



import os
import collections

from dl.models.keras_applications import imagenet_utils

from dl.models.classification_models import get_submodules_from_kwargs
from ._common_blocks import GroupConv2D, ChannelSE
from ..weights import load_model_weights

backend = None
layers = None
models = None
keras_utils = None

ModelParams = collections.namedtuple(
    'ModelParams',
    ['model_name', 'repetitions', 'residual_block', 'groups',
     'reduction', 'init_filters', 'input_3x3', 'dropout']
)


# -------------------------------------------------------------------------
#   Helpers functions
# -------------------------------------------------------------------------

def get_bn_params(**params):
    axis = 3 if backend.image_data_format() == 'channels_last' else 1
    default_bn_params = {
        'axis': axis,
        'epsilon': 9.999999747378752e-06,
    }
    default_bn_params.update(params)
    return default_bn_params


def get_num_channels(tensor):
    channels_axis = 3 if backend.image_data_format() == 'channels_last' else 1
    return backend.int_shape(tensor)[channels_axis]


# -------------------------------------------------------------------------
#   Residual blocks
# -------------------------------------------------------------------------

def SEResNetBottleneck(filters, reduction=16, strides=1, **kwargs):
    bn_params = get_bn_params()

    def layer(input_tensor):
        x = input_tensor
        residual = input_tensor

        # bottleneck
        x = layers.Conv2D(filters // 4, (1, 1), kernel_initializer='he_uniform',
                          strides=strides, use_bias=False)(x)
        x = layers.BatchNormalization(**bn_params)(x)
        x = layers.Activation('relu')(x)

        x = layers.ZeroPadding2D(1)(x)
        x = layers.Conv2D(filters // 4, (3, 3),
                          kernel_initializer='he_uniform', use_bias=False)(x)
        x = layers.BatchNormalization(**bn_params)(x)
        x = layers.Activation('relu')(x)

        x = layers.Conv2D(filters, (1, 1), kernel_initializer='he_uniform', use_bias=False)(x)
        x = layers.BatchNormalization(**bn_params)(x)

        #  if number of filters or spatial dimensions changed
        #  make same manipulations with residual connection
        x_channels = get_num_channels(x)
        r_channels = get_num_channels(residual)

        if strides != 1 or x_channels != r_channels:
            residual = layers.Conv2D(x_channels, (1, 1), strides=strides,
                                     kernel_initializer='he_uniform', use_bias=False)(residual)
            residual = layers.BatchNormalization(**bn_params)(residual)

        # apply attention module
        x = ChannelSE(reduction=reduction, **kwargs)(x)

        # add residual connection
        x = layers.Add()([x, residual])

        x = layers.Activation('relu')(x)

        return x

    return layer


def SEResNeXtBottleneck(filters, reduction=16, strides=1, groups=32, base_width=4, **kwargs):
    bn_params = get_bn_params()

    def layer(input_tensor):
        x = input_tensor
        residual = input_tensor

        width = (filters // 4) * base_width * groups // 64

        # bottleneck
        x = layers.Conv2D(width, (1, 1), kernel_initializer='he_uniform', use_bias=False)(x)
        x = layers.BatchNormalization(**bn_params)(x)
        x = layers.Activation('relu')(x)

        x = layers.ZeroPadding2D(1)(x)
        x = GroupConv2D(width, (3, 3), strides=strides, groups=groups,
                        kernel_initializer='he_uniform', use_bias=False, **kwargs)(x)
        x = layers.BatchNormalization(**bn_params)(x)
        x = layers.Activation('relu')(x)

        x = layers.Conv2D(filters, (1, 1), kernel_initializer='he_uniform', use_bias=False)(x)
        x = layers.BatchNormalization(**bn_params)(x)

        #  if number of filters or spatial dimensions changed
        #  make same manipulations with residual connection
        x_channels = get_num_channels(x)
        r_channels = get_num_channels(residual)

        if strides != 1 or x_channels != r_channels:
            residual = layers.Conv2D(x_channels, (1, 1), strides=strides,
                                     kernel_initializer='he_uniform', use_bias=False)(residual)
            residual = layers.BatchNormalization(**bn_params)(residual)

        # apply attention module
        x = ChannelSE(reduction=reduction, **kwargs)(x)

        # add residual connection
        x = layers.Add()([x, residual])

        x = layers.Activation('relu')(x)

        return x

    return layer


def SEBottleneck(filters, reduction=16, strides=1, groups=64, is_first=False, **kwargs):
    bn_params = get_bn_params()
    modules_kwargs = ({k: v for k, v in kwargs.items()
                       if k in ('backend', 'layers', 'models', 'utils')})

    if is_first:
        downsample_kernel_size = (1, 1)
        padding = False
    else:
        downsample_kernel_size = (3, 3)
        padding = True

    def layer(input_tensor):

        x = input_tensor
        residual = input_tensor

        # bottleneck
        x = layers.Conv2D(filters // 2, (1, 1), kernel_initializer='he_uniform', use_bias=False)(x)
        x = layers.BatchNormalization(**bn_params)(x)
        x = layers.Activation('relu')(x)

        x = layers.ZeroPadding2D(1)(x)
        x = GroupConv2D(filters, (3, 3), strides=strides, groups=groups,
                        kernel_initializer='he_uniform', use_bias=False, **kwargs)(x)
        x = layers.BatchNormalization(**bn_params)(x)
        x = layers.Activation('relu')(x)

        x = layers.Conv2D(filters, (1, 1), kernel_initializer='he_uniform', use_bias=False)(x)
        x = layers.BatchNormalization(**bn_params)(x)

        #  if number of filters or spatial dimensions changed
        #  make same manipulations with residual connection
        x_channels = get_num_channels(x)
        r_channels = get_num_channels(residual)

        if strides != 1 or x_channels != r_channels:
            if padding:
                residual = layers.ZeroPadding2D(1)(residual)
            residual = layers.Conv2D(x_channels, downsample_kernel_size, strides=strides,
                                     kernel_initializer='he_uniform', use_bias=False)(residual)
            residual = layers.BatchNormalization(**bn_params)(residual)

        # apply attention module
        x = ChannelSE(reduction=reduction, **kwargs)(x)

        # add residual connection
        x = layers.Add()([x, residual])

        x = layers.Activation('relu')(x)

        return x

    return layer


# -------------------------------------------------------------------------
#   SeNet builder
# -------------------------------------------------------------------------


def SENet(
        model_params,
        input_tensor=None,
        input_shape=None,
        include_top=True,
        classes=1000,
        weights='imagenet',
        **kwargs
):
    """Instantiates the ResNet, SEResNet architecture.
    Optionally loads weights pre-trained on ImageNet.
    Note that the data format convention used by the model is
    the one specified in your Keras config at `~/.keras/keras.json`.

    Args:
        include_top: whether to include the fully-connected
            layer at the top of the network.
        weights: one of `None` (random initialization),
              'imagenet' (pre-training on ImageNet),
              or the path to the weights file to be loaded.
        input_tensor: optional Keras tensor
            (i.e. output of `layers.Input()`)
            to use as image input for the model.
        input_shape: optional shape tuple, only to be specified
            if `include_top` is False (otherwise the input shape
            has to be `(224, 224, 3)` (with `channels_last` data format)
            or `(3, 224, 224)` (with `channels_first` data format).
            It should have exactly 3 inputs channels.
        classes: optional number of classes to classify images
            into, only to be specified if `include_top` is True, and
            if no `weights` argument is specified.

    Returns:
        A Keras model instance.

    Raises:
        ValueError: in case of invalid argument for `weights`,
            or invalid input shape.
    """

    global backend, layers, models, keras_utils
    backend, layers, models, keras_utils = get_submodules_from_kwargs(kwargs)

    residual_block = model_params.residual_block
    init_filters = model_params.init_filters
    bn_params = get_bn_params()

    # define input
    if input_tensor is None:
        input = layers.Input(shape=input_shape, name='input')
    else:
        if not backend.is_keras_tensor(input_tensor):
            input = layers.Input(tensor=input_tensor, shape=input_shape)
        else:
            input = input_tensor

    x = input

    if model_params.input_3x3:

        x = layers.ZeroPadding2D(1)(x)
        x = layers.Conv2D(init_filters, (3, 3), strides=2,
                          use_bias=False, kernel_initializer='he_uniform')(x)
        x = layers.BatchNormalization(**bn_params)(x)
        x = layers.Activation('relu')(x)

        x = layers.ZeroPadding2D(1)(x)
        x = layers.Conv2D(init_filters, (3, 3), use_bias=False,
                          kernel_initializer='he_uniform')(x)
        x = layers.BatchNormalization(**bn_params)(x)
        x = layers.Activation('relu')(x)

        x = layers.ZeroPadding2D(1)(x)
        x = layers.Conv2D(init_filters * 2, (3, 3), use_bias=False,
                          kernel_initializer='he_uniform')(x)
        x = layers.BatchNormalization(**bn_params)(x)
        x = layers.Activation('relu')(x)

    else:
        x = layers.ZeroPadding2D(3)(x)
        x = layers.Conv2D(init_filters, (7, 7), strides=2, use_bias=False,
                          kernel_initializer='he_uniform')(x)
        x = layers.BatchNormalization(**bn_params)(x)
        x = layers.Activation('relu')(x)

    x = layers.ZeroPadding2D(1)(x)
    x = layers.MaxPooling2D((3, 3), strides=2)(x)

    # body of resnet
    filters = model_params.init_filters * 2
    for i, stage in enumerate(model_params.repetitions):

        # increase number of filters with each stage
        filters *= 2

        for j in range(stage):

            # decrease spatial dimensions for each stage (except first, because we have maxpool before)
            if i == 0 and j == 0:
                x = residual_block(filters, reduction=model_params.reduction,
                                   strides=1, groups=model_params.groups, is_first=True, **kwargs)(x)

            elif i != 0 and j == 0:
                x = residual_block(filters, reduction=model_params.reduction,
                                   strides=2, groups=model_params.groups, **kwargs)(x)
            else:
                x = residual_block(filters, reduction=model_params.reduction,
                                   strides=1, groups=model_params.groups, **kwargs)(x)

    if include_top:
        x = layers.GlobalAveragePooling2D()(x)
        if model_params.dropout is not None:
            x = layers.Dropout(model_params.dropout)(x)
        x = layers.Dense(classes)(x)
        x = layers.Activation('softmax', name='output')(x)

    model = models.Model(input, x)

    if weights:
        if type(weights) == str and os.path.exists(weights):
            model.load_weights(weights)
        else:
            load_model_weights(model, model_params.model_name,
                               weights, classes, include_top, **kwargs)

    return model


# -------------------------------------------------------------------------
#   SE Residual Models
# -------------------------------------------------------------------------

MODELS_PARAMS = {
    'seresnet50': ModelParams(
        'seresnet50', repetitions=(3, 4, 6, 3), residual_block=SEResNetBottleneck,
        groups=1, reduction=16, init_filters=64, input_3x3=False, dropout=None,
    ),

    'seresnet101': ModelParams(
        'seresnet101', repetitions=(3, 4, 23, 3), residual_block=SEResNetBottleneck,
        groups=1, reduction=16, init_filters=64, input_3x3=False, dropout=None,
    ),

    'seresnet152': ModelParams(
        'seresnet152', repetitions=(3, 8, 36, 3), residual_block=SEResNetBottleneck,
        groups=1, reduction=16, init_filters=64, input_3x3=False, dropout=None,
    ),

    'seresnext50': ModelParams(
        'seresnext50', repetitions=(3, 4, 6, 3), residual_block=SEResNeXtBottleneck,
        groups=32, reduction=16, init_filters=64, input_3x3=False, dropout=None,
    ),

    'seresnext101': ModelParams(
        'seresnext101', repetitions=(3, 4, 23, 3), residual_block=SEResNeXtBottleneck,
        groups=32, reduction=16, init_filters=64, input_3x3=False, dropout=None,
    ),

    'senet154': ModelParams(
        'senet154', repetitions=(3, 8, 36, 3), residual_block=SEBottleneck,
        groups=64, reduction=16, init_filters=64, input_3x3=True, dropout=0.2,
    ),
}


def SEResNet50(input_shape=None, input_tensor=None, weights=None, classes=1000, include_top=True, **kwargs):
    return SENet(
        MODELS_PARAMS['seresnet50'],
        input_shape=input_shape,
        input_tensor=input_tensor,
        include_top=include_top,
        classes=classes,
        weights=weights,
        **kwargs
    )


def SEResNet101(input_shape=None, input_tensor=None, weights=None, classes=1000, include_top=True, **kwargs):
    return SENet(
        MODELS_PARAMS['seresnet101'],
        input_shape=input_shape,
        input_tensor=input_tensor,
        include_top=include_top,
        classes=classes,
        weights=weights,
        **kwargs
    )


def SEResNet152(input_shape=None, input_tensor=None, weights=None, classes=1000, include_top=True, **kwargs):
    return SENet(
        MODELS_PARAMS['seresnet152'],
        input_shape=input_shape,
        input_tensor=input_tensor,
        include_top=include_top,
        classes=classes,
        weights=weights,
        **kwargs
    )


def SEResNeXt50(input_shape=None, input_tensor=None, weights=None, classes=1000, include_top=True, **kwargs):
    return SENet(
        MODELS_PARAMS['seresnext50'],
        input_shape=input_shape,
        input_tensor=input_tensor,
        include_top=include_top,
        classes=classes,
        weights=weights,
        **kwargs
    )


def SEResNeXt101(input_shape=None, input_tensor=None, weights=None, classes=1000, include_top=True, **kwargs):
    return SENet(
        MODELS_PARAMS['seresnext101'],
        input_shape=input_shape,
        input_tensor=input_tensor,
        include_top=include_top,
        classes=classes,
        weights=weights,
        **kwargs
    )


def SENet154(input_shape=None, input_tensor=None, weights=None, classes=1000, include_top=True, **kwargs):
    return SENet(
        MODELS_PARAMS['senet154'],
        input_shape=input_shape,
        input_tensor=input_tensor,
        include_top=include_top,
        classes=classes,
        weights=weights,
        **kwargs
    )


def preprocess_input(x, **kwargs):
    return imagenet_utils.preprocess_input(x, mode='torch', **kwargs)


setattr(SEResNet50, '__doc__', SENet.__doc__)
setattr(SEResNet101, '__doc__', SENet.__doc__)
setattr(SEResNet152, '__doc__', SENet.__doc__)
setattr(SEResNeXt50, '__doc__', SENet.__doc__)
setattr(SEResNeXt101, '__doc__', SENet.__doc__)
setattr(SENet154, '__doc__', SENet.__doc__)



import functools
import dl.models.keras_applications as ka

from .models import resnet as rn
from .models import resnext as rx
from .models import senet as sn


class ModelsFactory:
    _models = {

        # ResNets
        'resnet18': [rn.ResNet18, rn.preprocess_input],
        'resnet34': [rn.ResNet34, rn.preprocess_input],
        'resnet50': [rn.ResNet50, rn.preprocess_input],
        'resnet101': [rn.ResNet101, rn.preprocess_input],
        'resnet152': [rn.ResNet152, rn.preprocess_input],

        # SE-Nets
        'seresnet18': [rn.SEResNet18, rn.preprocess_input],
        'seresnet34': [rn.SEResNet34, rn.preprocess_input],
        'seresnet50': [sn.SEResNet50, sn.preprocess_input],
        'seresnet101': [sn.SEResNet101, sn.preprocess_input],
        'seresnet152': [sn.SEResNet152, sn.preprocess_input],
        'seresnext50': [sn.SEResNeXt50, sn.preprocess_input],
        'seresnext101': [sn.SEResNeXt101, sn.preprocess_input],
        'senet154': [sn.SENet154, sn.preprocess_input],

        # Resnet V2
        'resnet50v2': [ka.resnet_v2.ResNet50V2, ka.resnet_v2.preprocess_input],
        'resnet101v2': [ka.resnet_v2.ResNet101V2, ka.resnet_v2.preprocess_input],
        'resnet152v2': [ka.resnet_v2.ResNet152V2, ka.resnet_v2.preprocess_input],

        # ResNext
        'resnext50': [rx.ResNeXt50, rx.preprocess_input],
        'resnext101': [rx.ResNeXt101, rx.preprocess_input],

        # VGG
        'vgg16': [ka.vgg16.VGG16, ka.vgg16.preprocess_input],
        'vgg19': [ka.vgg19.VGG19, ka.vgg19.preprocess_input],

        # Densnet
        'densenet121': [ka.densenet.DenseNet121, ka.densenet.preprocess_input],
        'densenet169': [ka.densenet.DenseNet169, ka.densenet.preprocess_input],
        'densenet201': [ka.densenet.DenseNet201, ka.densenet.preprocess_input],

        # Inception
        'inceptionresnetv2': [ka.inception_resnet_v2.InceptionResNetV2,
                              ka.inception_resnet_v2.preprocess_input],
        'inceptionv3': [ka.inception_v3.InceptionV3, ka.inception_v3.preprocess_input],
        'xception': [ka.xception.Xception, ka.xception.preprocess_input],

        # Nasnet
        'nasnetlarge': [ka.nasnet.NASNetLarge, ka.nasnet.preprocess_input],
        'nasnetmobile': [ka.nasnet.NASNetMobile, ka.nasnet.preprocess_input],

        # MobileNet
        'mobilenet': [ka.mobilenet.MobileNet, ka.mobilenet.preprocess_input],
        'mobilenetv2': [ka.mobilenet_v2.MobileNetV2, ka.mobilenet_v2.preprocess_input],
    }

    @property
    def models(self):
        return self._models

    def models_names(self):
        return list(self.models.keys())

    @staticmethod
    def get_kwargs():
        return {}

    def inject_submodules(self, func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            modules_kwargs = self.get_kwargs()
            new_kwargs = dict(list(kwargs.items()) + list(modules_kwargs.items()))
            return func(*args, **new_kwargs)

        return wrapper

    def get(self, name):
        if name not in self.models_names():
            raise ValueError('No such model `{}`, available models: {}'.format(
                name, list(self.models_names())))

        model_fn, preprocess_input = self.models[name]
        model_fn = self.inject_submodules(model_fn)
        preprocess_input = self.inject_submodules(preprocess_input)
        return model_fn, preprocess_input



# -*- coding: utf-8 -*-
"""
Created on Thu Oct  8 16:32:23 2020

@author: Koerber
"""

from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *
from ui.style import styleForm
from ui.ui_utils import LabelledSpinBox


class Window(object):
    def setupUi(self, Form):
        Form.setWindowTitle('Object Counting Settings') 
        styleForm(Form)
        
        self.centralWidget = QWidget(Form)
         
        vlayout = QVBoxLayout(self.centralWidget)
        vlayout.setContentsMargins(3, 3, 3, 3)
        vlayout.setSpacing(2)
        

        hlayout = QHBoxLayout(self.centralWidget)
        self.SObjectSize =  QSlider(Qt.Horizontal, self.centralWidget)
        self.SObjectSize.setMinimum(1)
        self.SObjectSize.setMaximum(10)
        self.SObjectSize.setObjectName('nn_CountingObjectSize')
        self.SObjectSize.setToolTip('Use to adapt object size and distribution. Use smaller values for small and densely distributed objects')
        self.LObjSize = QLabel(self.centralWidget)
        self.LObjSize.setText('Object Size')
        hlayout.addWidget(self.SObjectSize)
        hlayout.addWidget(self.LObjSize)
        
        h2layout = QHBoxLayout(self.centralWidget)
        self.SObjectIntensity =  QSlider(Qt.Horizontal, self.centralWidget)
        self.SObjectIntensity.setMinimum(1)
        self.SObjectIntensity.setMaximum(500)
        self.SObjectIntensity.setObjectName('nn_CountingObjectIntensity')
        self.SObjectIntensity.setToolTip('Use to set object peak intensity')
        self.LObjIntensity = QLabel(self.centralWidget)
        self.LObjIntensity.setText('Object Intensity')

        self.CBIgnoreBackgroundTiles = QCheckBox("Prefer labelled parts",self.centralWidget)
        self.CBIgnoreBackgroundTiles.setToolTip('When checked image parts which contain only background are ignored during training')
        self.CBIgnoreBackgroundTiles.setObjectName('nn_IgnoreBackground_od')

        self.SBDiscard = LabelledSpinBox ('Discard up to x background tiles', self.centralWidget)
        self.SBDiscard.setObjectName('nn_DiscardBGTiles_od')
        self.SBDiscard.SpinBox.setRange(1,20)

        self.SBMinPixels = LabelledSpinBox ('Minimum required labelled pixels', self.centralWidget)
        self.SBMinPixels.setObjectName('nn_MinLabelledPixels_od')
        self.SBMinPixels.SpinBox.setRange(1,1000)


        
        h2layout.addWidget(self.SObjectIntensity)
        h2layout.addWidget(self.LObjIntensity)


        vlayout.addLayout(hlayout)
        vlayout.addLayout(h2layout)
        vlayout.addWidget(self.CBIgnoreBackgroundTiles)
        vlayout.addWidget(self.SBDiscard)
        vlayout.addWidget(self.SBMinPixels)


        self.centralWidget.setLayout(vlayout)
        self.centralWidget.setFixedSize(vlayout.sizeHint())
        Form.setFixedSize(vlayout.sizeHint())
                


class ObjectDetectionSettingsWindow(QMainWindow, Window):
    def __init__(self, parent ):
        super(ObjectDetectionSettingsWindow, self).__init__(parent)
        self.parent = parent
        self.setupUi(self)
        
        self.SObjectSize.valueChanged.connect(self.ObjectSizeChanged)
        self.SObjectIntensity.valueChanged.connect(self.ObjectIntensityChanged)

        self.CBIgnoreBackgroundTiles.stateChanged.connect(self.IgnoreBackgroundChanged)
        self.SBDiscard.SpinBox.valueChanged.connect(self.DiscardChanged)
        self.SBMinPixels.SpinBox.valueChanged.connect(self.MinPixelsChanged)
        
        self.CBIgnoreBackgroundTiles.setChecked(self.parent.parent.dl.augmentation.od_ignoreBackground)
        
    def show(self):
        self.SObjectSize.setValue(self.parent.parent.dl.od_kernel_stdev)
        self.SObjectIntensity.setValue(self.parent.parent.dl.od_peak_val)
        self.CBIgnoreBackgroundTiles.setChecked(self.parent.parent.dl.augmentation.od_ignoreBackground)
        self.SBDiscard.SpinBox.setValue(self.parent.parent.dl.augmentation.od_ignorecycles)
        self.SBMinPixels.SpinBox.setValue(self.parent.parent.dl.augmentation.od_minRequiredPixels)
        super(ObjectDetectionSettingsWindow, self).show()
        
    def ObjectSizeChanged(self):
        self.parent.parent.dl.od_kernel_stdev = self.SObjectSize.value()
        
    def ObjectIntensityChanged(self):
        self.parent.parent.dl.od_peak_val = self.SObjectIntensity.value()

    def IgnoreBackgroundChanged(self):
        self.parent.parent.dl.augmentation.od_ignoreBackground = self.CBIgnoreBackgroundTiles.isChecked()
        
        if self.CBIgnoreBackgroundTiles.isChecked():
            self.SBMinPixels.setEnabled(True)
            self.SBDiscard.setEnabled(True)
        else:
            self.SBMinPixels.setEnabled(False)
            self.SBDiscard.setEnabled(False)

    def MinPixelsChanged(self):
        self.parent.parent.dl.augmentation.od_minRequiredPixels = self.SBMinPixels.SpinBox.value()

    def DiscardChanged(self):
        self.parent.parent.dl.augmentation.od_ignorecycles = self.SBDiscard.SpinBox.value()
        
        
        


# this is a clone from https://github.com/qubvel/efficientnet
# with minor changes
# licence Apache 2.0, https://github.com/qubvel/efficientnet/blob/master/LICENSE


# Copyright 2019 The TensorFlow Authors, Pavel Yakubovskiy. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import functools
from .__version__ import __version__

_KERAS_BACKEND = None
_KERAS_LAYERS = None
_KERAS_MODELS = None
_KERAS_UTILS = None


def get_submodules_from_kwargs(kwargs):
    backend = kwargs.get('backend', _KERAS_BACKEND)
    layers = kwargs.get('layers', _KERAS_LAYERS)
    models = kwargs.get('models', _KERAS_MODELS)
    utils = kwargs.get('utils', _KERAS_UTILS)
    for key in kwargs.keys():
        if key not in ['backend', 'layers', 'models', 'utils']:
            raise TypeError('Invalid keyword argument: %s', key)
    return backend, layers, models, utils


def inject_keras_modules(func):
    import keras
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        kwargs['backend'] = keras.backend
        kwargs['layers'] = keras.layers
        kwargs['models'] = keras.models
        kwargs['utils'] = keras.utils
        return func(*args, **kwargs)

    return wrapper


def inject_tfkeras_modules(func):
    import tensorflow.keras as tfkeras
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        kwargs['backend'] = tfkeras.backend
        kwargs['layers'] = tfkeras.layers
        kwargs['models'] = tfkeras.models
        kwargs['utils'] = tfkeras.utils
        return func(*args, **kwargs)

    return wrapper


def init_keras_custom_objects():
    import keras
    from . import model

    custom_objects = {
        'swish': inject_keras_modules(model.get_swish)(),
        'FixedDropout': inject_keras_modules(model.get_dropout)()
    }

    keras.utils.generic_utils.get_custom_objects().update(custom_objects)


def init_tfkeras_custom_objects():
    import tensorflow.keras as tfkeras
    from . import model

    custom_objects = {
        'swish': inject_tfkeras_modules(model.get_swish)(),
        'FixedDropout': inject_tfkeras_modules(model.get_dropout)()
    }

    tfkeras.utils.get_custom_objects().update(custom_objects)
